# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

from contextlib import suppress, contextmanager
from . import _device, _traits, _rack
from ._device import Device
from ._traits import observe
from ._rack import Owner, Rack
from . import _host
from . import _device as core
from . import value
from . import util
import copy
import inspect
import io
import json
from numbers import Number
import os
from pathlib import Path
import pickle
import shutil
import tarfile
import warnings

EMPTY = inspect._empty

INSPECT_SKIP_FILES = _device.__file__, _traits.__file__, _rack.__file__, __file__


class MungerBase(core.Device):
    """Organize file output with a key in the root database.

    The following conversions to relational files are attempted in order to
    convert each value in the row dictionary:

    1. Text containing a valid file or directory *outside* of the root data
       directory is made relational by moving the file or directory into
       the current row. The text is replaced with the updated relative path;
    2. Text longer than `text_relational_min` is dumped into a relational
       text file;
    3. 1- or 2-D data is converted to a pandas Series or DataFrame, and
       dumped into a relational file defined by the extension set by
       `nonscalar_file_type`

    """

    resource = value.Path(help="base directory for all data")
    text_relational_min = value.int(
        1024,
        min=0,
        help="minimum size threshold that triggers storing text in a relational file",
    )
    force_relational = value.list(
        ["host_log"], help="list of column names to always save as relational data"
    )
    dirname_fmt = value.str(
        "{id} {host_time}",
        help="directory name format for data in each row keyed on column",
    )
    nonscalar_file_type = value.str(
        "csv", help="file format for non-scalar numerical data"
    )
    metadata_dirname = value.str("metadata", help="subdirectory name for metadata")

    def __call__(self, index, row):
        """
        Break special cases of row items that need to be stored in
        relational files. The row valueis replaced in-place with the relative
        path to the data saved on disk.

        Arguments:
            row: dictionary of {'entry_name': "entry_value"} pairs
        Returns:
            the row dictionary, replacing special entries with the relative path to the saved data file
        """

        def is_path(v):
            if not isinstance(v, str):
                return False
            try:
                return os.path.exists(v)
            except ValueError:
                return False

        for name, value in row.items():
            if is_path(value):
                # Path to a datafile to move into the dataset
                row[name] = self._from_external_file(name, value, index, row)

            elif isinstance(value, (str, bytes)):
                # A long string that should be written to a text file
                if (
                    len(value) > self.text_relational_min
                    or name in self.force_relational
                    or isinstance(value, bytes)
                ):
                    row[name] = self._from_text(name, value, index, row)

            elif isinstance(value, (np.ndarray, pd.Series, pd.DataFrame)):
                # vector, table, matrix, etc.
                row[name] = self._from_ndarraylike(name, value, index, row)

            elif hasattr(value, "__len__") or hasattr(value, "__iter__"):
                # tuple, list, or other iterable
                row[name] = self._from_sequence(name, value, index, row)

        return row

    @classmethod
    def __imports__(cls):
        global np, pd

        import numpy as np
        import pandas as pd

    def save_metadata(self, name, key_func, **extra):
        def process_value(value, key_name):
            if isinstance(value, (str, bytes)):
                if len(value) > self.text_relational_min:
                    self._from_text(key_name, value)
                else:
                    return value
            elif hasattr(value, "__len__") or hasattr(value, "__iter__"):
                if not hasattr(value, "__len__") or len(value) > 0:
                    self._from_ndarraylike(key_name, value)
                else:
                    return ""
            else:
                return value

        summary = dict(extra)
        for owner, owner_name in name.items():

            for trait_name, trait in owner._traits.items():
                if trait.role == _traits.Trait.ROLE_VALUE or trait.cache:
                    summary[key_func(owner_name, trait_name)] = getattr(
                        owner, trait_name
                    )
        summary = {k: process_value(v, k) for k, v in summary.items()}

        metadata = dict(summary=pd.DataFrame([summary], index=["Value"]).T)

        self._write_metadata(metadata)

    def _write_metadata(self, metadata):
        raise NotImplementedError

    def _from_ndarraylike(self, name, value, index=0, row=None):
        """Write nonscalar (potentially array-like, or a python object) data
        to a file, and return a path to the file

        Arguments:
            name: name of the entry to write, used as the filename
            value: the object containing array-like data
            row: row dictionary, or None (the default) to write to the metadata folder
        Returns:
            the path to the file, relative to the directory that contains the root database
        """

        def write(stream, ext, value):
            if ext == "csv":
                value.to_csv(stream)
            elif ext == "json":
                value.to_json(stream)
            elif ext in ("p", "pickle"):
                pickle.dump(value, stream, 2)
            elif ext == "feather":
                to_feather(value, stream)
            elif ext == "db":
                raise Exception("sqlite not implemented for relational files")
            else:
                raise Exception(f"extension {ext} doesn't match a known format")

        if hasattr(value, "__len__") and len(value) == 0:
            return ""

        if row is None:
            ext = "csv"
        else:
            ext = self.nonscalar_file_type

        try:
            value = pd.DataFrame(value)
            if value.shape[0] == 0:
                value = pd.DataFrame([value])
        except BaseException:
            # We couldn't make a DataFrame
            self._logger.error(
                f"Failed to form DataFrame from {repr(name)}; pickling object instead"
            )
            ext = "pickle"
        finally:
            if row is None:
                stream = self._open_metadata(name + "." + ext, "wb")
            else:
                stream = self._open_relational(name + "." + ext, index, row, mode="wb")

            # Workaround for bytes/str encoding quirk underlying pandas 0.23.1
            try:
                write(stream, ext, value)
            except TypeError:
                with io.TextIOWrapper(stream, newline="\n") as buf:
                    write(buf, ext, value)
            return self._get_key(stream)

    def _from_external_file(self, name, old_path, index=0, row=None, ntries=10):
        basename = os.path.basename(old_path)
        with self._open_relational(basename, index, row, mode="wb") as buf:
            new_path = buf.name

        self._import_from_file(old_path, new_path)

        return self._get_key(new_path)

    def _from_text(self, name, value, index=0, row=None, ext=".txt"):
        """Write a string data to a file

        Arguments:
            name: name of the parameter (helps to determine file path)
            value: the string to write to file
            row: the row to infer timestamp, or None to write to metadata
            ext: file extension
        Returns:
            the path to the file, relative to the directory that contains the root database
        """
        with self._open_relational(name + ext, index, row, "w") as f:
            f.write(value)
        return self._get_key(f)

    def _from_sequence(self, name, value, index=0, row=None, ext=".json"):
        """Write a string data to a file

        Arguments:
            name: name of the parameter (helps to determine file path)
            value: the string to write to file
            row: the row to infer timestamp, or None to write to metadata
            ext: file extension
        Returns:
            the path to the file, relative to the directory that contains the root database
        """
        with self._open_relational(name + ext, index, row, mode="w") as f:
            json.dump(value, f, indent=True)  # f.write(bytes(value, encoding='utf-8'))
        return self._get_key(f)

    # The following methods need to be implemented in subclasses.
    def _get_key(self, buf):
        """Key to use for the relative data in the root database?

        Arguments:
            stream: stream for writing to the relational data file
        Returns:
            the key
        """
        raise NotImplementedError

    def _open_relational(self, name, index, row, mode):
        """Open a stream / IO buffer for writing relational data, given
        the root database column, index, and the row dictionary.

        Arguments:
            name: the column name of the relational data in the root db
            index: the index name of of the data in the root db
            row: the dictionary containing the row of data at `index`

        Returns:

            an open buffer object for writing data
        """
        raise NotImplementedError

    def _open_metadata(self, name):
        """Open a stream / IO buffer for writing metadata, given
        the name of the metadata.

        Returns:

            an open buffer object for writing metadata to the file
        """

        raise NotImplementedError

    def _import_from_file(self, old_path, dest):
        raise NotImplementedError


class MungeToDirectory(MungerBase):
    """Implement data munging into subdirectories."""

    def _open_relational(self, name, index, row, mode):
        if "host_time" not in row:
            self._logger.error(
                "no timestamp yet from host yet; this shouldn't happen :("
            )

        relpath = self._make_path_heirarchy(index, row)
        if not os.path.exists(relpath):
            os.makedirs(relpath)

        return open(os.path.join(relpath, name), mode)

    def _open_metadata(self, name, mode):
        dirpath = os.path.join(self.resource, self.metadata_dirname)
        with suppress(FileExistsError):
            os.makedirs(dirpath)
        return open(os.path.join(dirpath, name), mode)

    def _get_key(self, stream):
        """Key to use for the relative data in the root database?

        Arguments:
            stream: stream for writing to the relational data file
        Returns:
            the key
        """
        return os.path.relpath(stream.name, self.resource)

    def _import_from_file(self, old_path, dest):
        # Retry moves for several seconds in case the file is in the process
        # of being closed
        @util.until_timeout(PermissionError, 5, delay=0.5)
        def move():
            os.rename(old_path, dest)

        try:
            move()
            self._logger.debug(f"moved {repr(old_path)} to {repr(dest)}")
        except PermissionError:
            self._logger.warning(
                "relational file was still open in another program; fallback to copy instead of rename"
            )
            import shutil

            shutil.copyfile(old_path, dest)

    def _make_path_heirarchy(self, index, row):
        relpath = self.dirname_fmt.format(id=index, **row)

        # TODO: Find a more generic way to force valid path name
        relpath = relpath.replace(":", "")
        relpath = os.path.join(self.resource, relpath)
        return relpath

    def _write_metadata(self, metadata):
        def recursive_dict_fix(d):
            d = dict(d)
            for name, obj in dict(d).items():
                if isinstance(obj, Path):
                    d[name] = str(obj)
                elif isinstance(obj, bytes):
                    d[name] = obj.decode()
                elif isinstance(obj, dict):
                    d[name] = recursive_dict_fix(obj)
            return d

        for k, v in metadata.items():
            stream = self._open_metadata(k + ".json", "w")
            if hasattr(stream, "overwrite"):
                stream.overwrite = True

            if isinstance(v, pd.DataFrame):
                v = v.to_dict()["Value"]
            if isinstance(v, dict):
                v = recursive_dict_fix(v)

            # with io.TextIOWrapper(stream, newline='\n') as buf:
            json.dump(v, stream, indent=True, sort_keys=True)


class TarFileIO(io.BytesIO):
    """For appending data into new files in a tarfile"""

    def __init__(self, open_tarfile, relname, mode="w", overwrite=False):
        #        self.tarbase = tarbase
        #        self.tarname = tarname
        self.tarfile = open_tarfile
        self.overwrite = False
        self.name = relname
        self.mode = mode
        super(TarFileIO, self).__init__()

    def __del__(self):
        try:
            super(TarFileIO, self).close()
        except ValueError:
            pass

        super(TarFileIO, self).__del__()

    #    def write(self, data, encoding='ascii'):
    #        if isinstance(data, str):
    #            data = bytes(data, encoding=encoding)
    #        super(TarFileIO,self).write(data)
    #

    def close(self):
        # First: dump the data into the tar file
        #        tarpath = os.path.join(self.tarbase, self.tarname)
        #        f = tarfile.open(tarpath, 'a')

        try:
            if not self.overwrite and self.name in self.tarfile.getnames():
                raise IOError(f"{self.name} already exists in {self.tarfile.name}")

            tarinfo = tarfile.TarInfo(self.name)
            tarinfo.size = self.tell()

            self.seek(0)

            self.tarfile.addfile(tarinfo, self)

        # Then make sure to close everything
        finally:
            super(TarFileIO, self).close()


class MungeToTar(MungerBase):
    """Implement data munging into a tar file. This is slower than
    MungeToDirectory but is tidier on the filesystem.
    """

    tarname = "data.tar"

    def _open_relational(self, name, index, row, mode):
        if "host_time" not in row:
            self._logger.error(
                "no timestamp yet from host yet; this shouldn't happen :("
            )

        relpath = os.path.join(self.dirname_fmt.format(id=index, **row), name)

        return TarFileIO(self.tarfile, relpath, mode=mode)

    def _open_metadata(self, name, mode):
        dirpath = os.path.join(self.metadata_dirname, name)
        return TarFileIO(self.tarfile, dirpath, mode=mode)

    def open(self):

        if not os.path.exists(self.resource):
            with suppress(FileExistsError):
                os.makedirs(self.resource)
        self.tarfile = tarfile.open(os.path.join(self.resource, self.tarname), "a")

    def close(self):
        util.logger.warning("MungeToTar cleanup()")
        self.tarfile.close()

    def _get_key(self, buf):
        """Where is the file relative to the root database?

        Arguments:
            path: path of the file
        Returns:
            path to the file relative to the root database
        """
        return buf.name

    def _import_from_file(self, old_path, dest):
        info = self.tar.gettarinfo(old_path)
        info.name = dest
        self.tar.addfile(old_path, info)

        @util.until_timeout(PermissionError, 5, delay=0.5)
        def remove():
            try:
                shutil.rmtree(old_path)
            except NotADirectoryError:
                os.remove(old_path)

        try:
            remove()
            self._logger.debug(f"moved {old_path} to into tar file as {dest}")
        except PermissionError:
            self._logger.warning(f"could not remove old file or directory {old_path}")

    def _write_metadata(self, metadata):
        for k, v in metadata.items():
            stream = self._open_metadata(k + ".json", "w")
            if hasattr(stream, "overwrite"):
                stream.overwrite = True

            if isinstance(v, pd.DataFrame):
                v = v.to_dict()["Value"]
            if isinstance(v, dict):
                for name, obj in v.items():
                    if isinstance(obj, Path):
                        v[name] = str(obj)

            with io.TextIOWrapper(stream, newline="\n") as buf:
                json.dump(v, buf, indent=True, sort_keys=True)


class Aggregator(util.Ownable):
    """Passive aggregation of data from Device property trait and value traits traits, and from calls to methods in Rack instances"""

    PERSISTENT_TRAIT_ROLES = (_traits.Trait.ROLE_VALUE,)

    def __init__(self):
        # registry of names to use for trait owners
        self.name_map = {}
        self.trait_rules = dict(always={}, never={})

        # pending data
        self._pending_traits_volatile = (
            {}
        )  # pending = dict(state={}, value traits={}, rack={})
        self._pending_traits_persistent = {}
        self._pending_rack_output = {}
        self._pending_rack_input = {}

        self._rack_toplevel_caller = None
        self._rack_input_index = 0

        # self._iter_index_names = {}

        # cached data
        self.metadata = {}

        super().__init__()

    def __repr__(self):
        return f"{self.__class__.__qualname__}()"

    def enable(self):
        # catch return data as well
        _rack.notify.observe_returns(self._receive_rack_output)
        _rack.notify.observe_calls(self._receive_rack_input)
        # _rack.notify.observe_call_iteration(self._receive_rack_iteration)

    def disable(self):
        _rack.notify.unobserve_returns(self._receive_rack_output)
        _rack.notify.unobserve_calls(self._receive_rack_input)
        # _rack.notify.unobserve_call_iteration(self._receive_rack_input)

    def is_persistent_trait(self, device, attr):
        if not isinstance(device, core.Device):
            return False

        trait = device._traits[attr]

        return trait.role in self.PERSISTENT_TRAIT_ROLES or trait.cache

    def get(self) -> list([dict, dict]):
        """return an aggregated dictionary output data (from Device traits and Rack method returns)
        and an aggregated dictionary of input data (keyword arguments passed to the toplevel Rack method).
        A get is
        also performed on each Device trait that is configured as "always" with `self.observe`, and any traits
        labeled "never" are removed.

        Returns:

            dictionary keyed on :func:`key` (defaults '{device name}_{state name}')
        """

        for device, name in list(self.name_map.items()):
            # Perform gets for each property trait called out in self.trait_rules['always']
            if device in self.trait_rules["always"].keys():
                for attr in self.trait_rules["always"][device]:
                    if self.is_persistent_trait(device, attr):
                        self._pending_traits_persistent[self.key(name, attr)] = getattr(
                            device, attr
                        )
                    else:
                        self._pending_traits_volatile[self.key(name, attr)] = getattr(
                            device, attr
                        )

            # Remove keys corresponding with self.trait_rules['never']
            if device in self.trait_rules["never"].keys():
                for attr in self.trait_rules["never"][device]:
                    if self.is_persistent_trait(device, attr):
                        self._pending_traits_persistent.pop(self.key(name, attr), None)
                    else:
                        self._pending_traits_volatile.pop(self.key(name, attr), None)

        aggregated_output = dict(index=self._rack_input_index)

        # start by aggregating the trait data, and checking for conflicts with keys in the Rack data
        aggregated_output.update(self._pending_traits_persistent)
        aggregated_output.update(self._pending_traits_volatile)

        # check namespace conflicts and pull in rack outputs
        key_conflicts = set(aggregated_output).intersection(self._pending_rack_output)
        if len(key_conflicts) > 0:
            self.critical(
                f"key name conflict in aggregated data - Rack data is overwriting trait data for {key_conflicts}"
            )
        aggregated_output.update(self._pending_rack_output)

        # and the rack inputs
        aggregated_input = dict(index=self._rack_input_index)
        aggregated_input.update(self._pending_rack_input)

        # clear Rack data, as well as property trait data if we don't assume it is consistent.
        # value traits traits are locally cached, so it is safe to keep them in the next step
        self._pending_rack_output = {}
        self._pending_traits_volatile = {}
        # self._pending_rack_input = {}

        return aggregated_output, aggregated_input

    def key(self, device_name, state_name):
        """Generate a name for a trait based on the names of
        a device and one of its states or value traits.
        """
        return f"{device_name}_{state_name}"

    def set_device_labels(self, **mapping):
        """Manually choose device name for a device instance.

        Arguments:
            mapping (dict): name mapping, formatted as {device_object: 'device name'}
        Returns:
            None
        """
        for label, device in list(mapping.items()):
            if isinstance(device, Aggregator):
                pass
            elif not isinstance(device, Device):
                raise ValueError(f"{device} is not an instance of Device")

        self.name_map.update([(v, k) for k, v in mapping.items()])

    def _receive_rack_output(self, msg: dict):
        """called by an owning Rack notifying that managed procedural steps have returned data"""
        # trait data or previous returned data may cause problems here. perhaps this should be an exception?

        row_data = msg["new"]

        key_conflicts = set(row_data).intersection(self._pending_rack_output)
        if len(key_conflicts) > 0:
            self._logger.warning(
                f"Rack call overwrites prior data with existing keys {key_conflicts}"
            )
        self._pending_rack_output.update(row_data)

    def _receive_rack_input(self, msg: dict):
        """called by an owning Rack notifying that managed procedural steps have returned data"""
        # trait data or previous returned data may cause problems here. perhaps this should be an exception?

        if self._rack_toplevel_caller is None:
            # take the first call as the top-level caller
            self._rack_toplevel_caller = msg['owner']

        elif self._rack_toplevel_caller != msg['owner']:
            # only track calls to the top-level caller
            return

        else:
            self._rack_input_index += 1

        iter_info = msg["new"]
        row_data = dict(index=self._rack_input_index)

        # if len(self._iter_index_names) > 0:
        #     self._iter_index_names.setdefault(
        #         msg["owner"], "index_" + msg["owner"]._owned_name
        #     )
        # else:
        #     self._iter_index_names.setdefault(msg["owner"], "index")

        # row_data[self._iter_index_names[msg["owner"]]] = iter_info["index"]
        # if iter_info['step_name']:
        #     row_data[msg['owner']._owned_name + '_step_name'] = iter_info['step_name']

        # TODO: should there be some kind of conflict check for this?
        # key_conflicts = set(row_data).intersection(self._pending_traits_persistent)
        # if len(key_conflicts) > 0:
        #     self._logger.warning(
        #         f"Rack call overwrites prior data with existing keys {key_conflicts}"
        #     )

        self._pending_rack_input = dict(row_data, **msg['new'])

    def _receive_trait_update(self, msg: dict):
        """called by trait owners on changes observed

        Arguments:
            msg (dict): callback info dictionary generated by traitlets
        Returns:
            None
        """

        if msg["name"].startswith("_"):
            # skip private traits
            return

        name = self.name_map[msg["owner"]]
        attr = msg["name"]

        if msg["cache"]:
            self.metadata[self.key(name, attr)] = msg["new"]
        elif self.is_persistent_trait(msg["owner"], msg["name"]):
            self._pending_traits_persistent[self.key(name, attr)] = msg["new"]
        else:
            self._pending_traits_volatile[self.key(name, attr)] = msg["new"]

    def _update_name_map(self, rack_or_devices):
        """ "map each Device to a name in devices.values() by introspection"""

        # self.name_map.clear()

        if isinstance(rack_or_devices, Rack):
            devices = {
                name: obj
                for name, obj in Rack._ownables.items()
                if isinstance(obj, Device)
            }
        else:
            devices = rack_or_devices

        for device, name in devices.items():
            if not isinstance(device, Device):
                raise ValueError(f"{device} is not an instance of Device")

            if name:
                if device in self.name_map and name != self.name_map[device]:
                    # a rename is an odd case, make a note of it
                    self._logger.warning(f"renaming {self.name_map[device]} to {name}")
                self.name_map[device] = name

            elif device in self.name_map:
                # naming for device. needs this
                name = self.name_map[device]

            elif hasattr(device, "__name__"):
                # Ownable instances that have owners have this attribute set
                self.name_map[device] = name = device.__name__

            else:
                self.name_map[device] = name = self._find_object_in_callers(device)
                self._logger.info(f"{device} named '{name}' by introspection")

        if len(list(self.name_map.values())) != len(set(self.name_map.values())):
            names = list(self.name_map.values())
            duplicates = set([x for x in names if names.count(x) > 1])

            raise Exception(
                f"could not automatically resolve duplicate device name(s) {duplicates}"
            )

    def observe(self, devices, changes=True, always=[], never=["isopen"]):
        """Configure the data to aggregate from value, property, or datareturn traits in the given devices.

        Device may be a single device instance, or an
        several devices in an iterable (such as a list
        or tuple) to apply to each one.

        Subsequent calls to :func:`observe_states` replace
        the existing list of observed property traits for each
        device.

        Arguments:
            devices: Device instance, iterable of Devices, or {device:name} mapping
            changes (bool): Whether to automatically log each time a property trait is set for the supplied device(s)
            always: name (or iterable of multiple names) of property traits to actively update on each call to get()
            never: name (or iterable of multiple names) of property traits to exclude from aggregated result (overrides :param:`always`)
        """

        # parameter checks
        if isinstance(devices, Device):
            devices = {devices: None}
        elif isinstance(devices, dict):
            pass
        elif hasattr(devices, "__iter__"):
            devices = dict([(d, None) for d in devices])
        else:
            raise ValueError("devices argument must be a device or iterable of devices")

        if isinstance(always, str):
            always = (always,)
        elif hasattr(always, "__iter__"):
            always = tuple(always)
        else:
            raise ValueError(
                "argument 'always' must be a str or iterable of str"
            )

        if isinstance(never, str):
            never = (never,)
        elif hasattr(never, "__iter__"):
            never = tuple(never)
        else:
            raise ValueError(
                "argument 'never' must be a str or iterable of str"
            )

        # if isinstance(role, (str,bytes)):
        #     role = [role]
        # TODO: remove this for good?
        # if role not in VALID_TRAIT_ROLES:
        #     raise ValueError(f"the 'role' argument must be one of {str(VALID_TRAIT_ROLES)}, not {role}")

        self._update_name_map(devices)

        # Register handlers
        for device in devices.keys():
            if changes:
                observe(device, self._receive_trait_update)
            else:
                core.unobserve(device, self._receive_trait_update)
            if always:
                self.trait_rules["always"][device] = always
            if never:
                self.trait_rules["never"][device] = never

    def _find_object_in_callers(self, target, max_levels=5):
        """Introspect into the caller to name an object .

        Arguments:
            target: device instance
            min_levels: minimum number of layers of calls to traverse
            max_levels: maximum number of layers to traverse
        Returns:
            name of the Device
        """

        #        # If the context is a function, look in its first argument,
        #        # in case it is a method. Search its class instance.
        #        if len(ret) == 0 and len(f.frame.f_code.co_varnames)>0:
        #            obj = f.frame.f_locals[f.frame.f_code.co_varnames[0]]
        #            for k, v in obj.__dict__.items():
        #                if isinstance(v, Device):
        #                    ret[k] = v

        def find_value(haystack, needle, reject=["self"]):
            for k, v in haystack.items():
                if v is needle:
                    if k in reject:
                        raise KeyError
                    else:
                        return k
            else:
                raise KeyError

        f = frame = inspect.currentframe()

        ret = None

        try:
            while f.f_code.co_filename in INSPECT_SKIP_FILES:
                f = f.f_back

            for i in range(max_levels):
                # look in the namespace; if nothing
                try:
                    ret = find_value(f.f_locals, target)
                except KeyError:
                    continue
                else:
                    break

                f = f.f_back
            else:
                raise Exception(f"failed to automatically label {repr(target)}")
        finally:
            del f, frame

        return ret


class RelationalTableLogger(
    Owner, util.Ownable, entry_order=(_host.Email, MungerBase, _host.Host)
):
    """Base class for loggers that queue dictionaries of data before writing
    to disk. This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path (str): Base path to use for the root database
        overwrite (bool): Whether to overwrite the root database if it exists (otherwise, append)

        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database

    Arguments:
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data
        metadata_dirname: The name of the subdirectory that should be used to store metadata (device connection parameters, etc.)

        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
        git_commit_in: perform a git commit on open() if the current
        directory is inside a git repo with this branch name
    """

    index_label = "id"

    def __init__(
        self,
        path=None,
        *,
        append=False,
        text_relational_min=1024,
        force_relational=["host_log"],
        dirname_fmt="{id} {host_time}",
        nonscalar_file_type="csv",
        metadata_dirname="metadata",
        tar=False,
        git_commit_in=None,
        # **metadata
    ):

        self.aggregator = Aggregator()

        super().__init__()

        # log host introspection
        # TODO: smarter
        self.host = _host.Host(git_commit_in=git_commit_in)

        # select the backend that dumps relational data
        munge_cls = MungeToTar if tar else MungeToDirectory
        self.munge = munge_cls(
            path,
            text_relational_min=text_relational_min,
            force_relational=force_relational,
            dirname_fmt=dirname_fmt,
            nonscalar_file_type=nonscalar_file_type,
            metadata_dirname=metadata_dirname,
            # **metadata
        )

        self.last_row = 0
        self.pending_output = []
        self.pending_input = []
        self.path = Path(path)
        self._append = append
        self.set_row_preprocessor(None)

    def __copy__(self):
        return copy.deepcopy(self)

    def __owner_init__(self, owner):
        super().__owner_init__(owner)

        devices, _ = _rack.recursive_devices(owner)

        self.observe({v: k for k, v in devices.items()})

    def __repr__(self):
        if hasattr(self, "path"):
            path = repr(str(self.path))
        else:
            path = "<unset path>"
        return f"{self.__class__.__qualname__}({path})"

    def observe(self, devices, changes=True, always=[], never=["isopen"]):
        """Configure the data to aggregate from value, property, or datareturn traits in the given devices.

        Device may be a single device instance, or an
        several devices in an iterable (such as a list
        or tuple) to apply to each one.

        Subsequent calls to :func:`observe_states` replace
        the existing list of observed property traits for each
        device.

        Arguments:
            devices: Device instance or iterable of Device instances
            changes (bool): Whether to automatically log each time a property trait is set for the supplied device(s)

            always: name (or iterable of multiple names) of property traits to actively update on each call to get()
            never: name (or iterable of multiple names) of property traits to exclude from aggregated result (overrides :param:`always`)
        """

        self.aggregator.observe(
            devices=devices, changes=changes, always=always, never=never
        )

    def set_row_preprocessor(self, func):
        """Define a function that is called to modify each pending data row
        before it is committed to disk. It should accept a single argument,
        a function or other callable that accepts a single argument (the
        row dictionary) and returns the dictionary modified for write
        to disk.
        """
        if func is None:
            self._row_preprocessor = lambda x: x
        elif callable(func):
            self._row_preprocessor = func
        else:
            raise ValueError("func argument must be callable or None")

    def new_row(self, *args, **kwargs):
        """Add a new row of data to the list of rows that await writes
        to disk, `self.pending_output`. This cache of pending data row is in
        the dictionary `self.pending_output`.

        The data row is represented as a dict in the form {'column_name': value}.
        This dict is formed here with
        * the most recent value, property, and datareturn traits ({trait_name: latest_value} as configured by `self.observe`);
        * optional positional dict args[0] (each {key: value});
        * optional keyword arguments in `kwargs` (each {key: value})

        Keyword arguments are passed in as ::

        >>>    db.new_row(dict(name0=value0), name1=value1, name2=value2, nameN=valueN)

        Simple "scalar" database types like numbers, booleans, and strings
        are added directly to the table. Non-scalar or multidimensional values are stored
        in a separate file (as defined in :func:`set_path_format`), and the path
        to this file is stored in the table.

        In order to write `self.pending_output` to disk, use :func:`self.write`.

        :param bool copy=True: When `True` (the default), use a deep copy of `data` to avoid
        problems with overwriting references to data if `data` is reused during test. This takes some extra time; set to `False` to skip this copy operation.

        Returns:
            the dictionary representation of the row added to `self.pending_output`.
        """
        do_copy = kwargs.pop("copy", None)

        # Start with input arguments
        if len(args) == 1:
            #            if not isinstance(args[0], dict):
            #                raise TypeError('argument to append must be a dictionary')
            row = copy.deepcopy(dict(args[0])) if do_copy else dict(args[0])
        elif len(args) == 0:
            row = {}

        # Pull in observed states
        aggregated_output, aggregated_input = self.aggregator.get()
        if do_copy:
            aggregated_output = copy.deepcopy(aggregated_output)
        row.update(aggregated_output)

        # Pull in keyword arguments
        if do_copy:
            kwargs = copy.deepcopy(dict(kwargs))
        else:
            kwargs = dict(kwargs)
        row.update(kwargs)

        self._logger.debug(f"new data row has {len(row)} columns")

        self.pending_output.append(row)
        self.pending_input.append(aggregated_input)

    def write(self):
        """Commit any pending rows to the root database, converting
        non-scalar data to data files, and replacing their dictionary value
        with the relative path to the data file.

        Returns:

            None
        """
        if len(self.pending_output) != len(self.pending_input):
            util.logger.warning('the input and output have mismatched length')

        count = max(len(self.pending_output), len(self.pending_input))

        if count > 0:
            proc = self._row_preprocessor

            self.pending_output = [
                self.munge(self.output_index + i, proc(row))
                for i, row in enumerate(self.pending_output)
            ]


            self.pending_input = [
                self.munge(self.output_index + i, proc(row))
                for i, row in enumerate(self.pending_input)
            ]

            self.output_index += len(self.pending_output) - 1
            self._write_root()
            self.clear()

    @contextmanager
    @util.hide_in_traceback
    def context(self, *args, **kws):
        """ calls `self.new_row(*args, **kws); self.write()` on context exit.

        This is meant as a convenience for defining execution behavior in
        table inputs for Racks.
        """

        try:
            yield self
        finally:
            self.new_row(*args, **kws)
            self.write()

    def _write_root(self):
        """Write pending data. This must be implemented by inheriting classes.

        Returns:
            None
        """
        raise NotImplementedError

    def clear(self):
        """Remove any queued data that has been added by append."""
        self.pending_output = []
        self.pending_input = []

    def set_relational_file_format(self, format):
        """Set the format to use for relational data files.

        Arguments:
            format (str): one of 'csv', 'json', 'feather', or 'pickle'
        """
        warnings.warn(
            """set_nonscalar_file_type is deprecated; set when creating
                         the database object instead with the nonscalar_output flag"""
        )

        if format not in ("csv", "json", "feather", "pickle", "db"):
            raise Exception(f"relational file data format {format} not supported")
        self.munge.nonscalar_file_type = format

    def set_path_format(self, format):
        """ Set the path name convention for relational files that is used when
            a table entry contains non-scalar (multidimensional) information and will
            need to be stored in a separate file. The entry in the aggregate property traits table
            becomes the path to the file.

            The format string follows the syntax of python's python's built-in :func:`str.format`.\
            You may use any keys from the table to form the path. For example, consider a\
            scenario where aggregate device property traits includes `inst1_frequency` of `915e6`,\
            and :func:`append` has been called as `append(dut="DUT15")`. If the current\
            aggregate property trait entry includes inst1_frequency=915e6, then the format string\
            '{dut}/{inst1_frequency}' means relative data path 'DUT15/915e6'.

            Arguments:
                format: a string compatible with :func:`str.format`, with replacement\
            fields defined from the keys from the current entry of results and aggregated states.\

            Returns:
                None
        """
        warnings.warn(
            """set_path_format is deprecated; set when creating
                         the database object instead with the nonscalar_output flag"""
        )

        self.munge.dirname_fmt = format

    def open(self):
        """Open the file or database connection.
        This is an abstract base method (to be overridden by inheriting classes)

        Returns:
            None
        """

        if self.path is None:
            raise TypeError(f"cannot open dB while path is None")

        self.observe(self.munge, never=self.munge._traits)
        self.observe(self.host, always=["time", "log"])

        # Do some checks on the relative data directory before we consider overwriting
        # the root db.
        self.aggregator.enable()

        self.clear()

        self.output_index = 0
        self._logger.debug(f"{self} is open")
        return self

    def close(self):
        self.aggregator.disable()
        # try:
        self.write()
        if self.output_index > 0:
            self.munge.save_metadata(
                self.aggregator.name_map,
                self.aggregator.key,
                **self.aggregator.metadata,
            )
        # except BaseException as e:
        #     traceback.print_exc()


class CSVLogger(RelationalTableLogger):
    """Store data, value traits, and property traits to disk into a root database formatted as a comma-separated value (CSV) file.

    This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path (str): Base path to use for the root database
        overwrite (bool): Whether to overwrite the root database if it exists (otherwise, append)
        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data
        metadata_dirname: The name of the subdirectory that should be used to store metadata (device connection parameters, etc.)
        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
    """

    ROOT_FILE_NAME = OUTPUT_FILE_NAME = "outputs.csv"
    INPUT_FILE_NAME = "inputs.csv"
    output_index = 0
    tables = {}

    nonscalar_file_type = "csv"

    def open(self):
        """Instead of calling `open` directly, consider using
        `with` statements to guarantee proper disconnection
        if there is an error. For example, the following
        sets up a connected instance::

            with CSVLogger('my.csv') as db:
                ### do the data acquisition here
                pass

        would instantiate a `CSVLogger` instance, and also guarantee
        a final attempt to write unwritten data is written, and that
        the file is closed when exiting the `with` block, even if there
        is an exception.
        """
        self.tables = {}

        self.path.mkdir(parents=True, exist_ok=self._append)

        for file_name in self.OUTPUT_FILE_NAME, self.INPUT_FILE_NAME:
            file_path = self.path / file_name
            try:
                # test access by starting the root table
                file_path.touch(exist_ok=self._append)
            except FileExistsError:
                raise IOError(
                    f"root table already exists at '{file_path}', while append=False"
                )

            if self._append and file_path.stat().st_size > 0:
                # there's something here and we plan to append
                self.tables[file_name] = pd.read_csv(file_path, nrows=1)
                self.tables[file_name].index.name = self.index_label
                self.output_index = len(self.tables[file_name])
            else:
                self.tables[file_name] = None
                self.output_index = 0

    def close(self):
        self._write_root()

    def _write_root(self):
        """Write queued rows of data to csv. This is called automatically on :func:`close`, or when
        exiting a `with` block.

        If the class was created with overwrite=True, then the first call to _write_root() will overwrite
        the preexisting file; subsequent calls append.
        """

        def append_csv(path_to_csv, df):
            if len(df) == 0:
                return
            isfirst = self.tables.get(path_to_csv.name,None) is None
            pending = pd.DataFrame(df)
            pending.index.name = self.index_label
            pending.index = pending.index + self.output_index

            if isfirst:
                self.tables[path_to_csv.name] = pending
            else:
                self.tables[path_to_csv.name] = self.tables[path_to_csv.name].append(pending).loc[self.output_index :]
            self.tables[path_to_csv.name].sort_index(inplace=True)
            self.output_index = self.tables[path_to_csv.name].index[-1]

            self.path.mkdir(exist_ok=True, parents=True)

            with open(path_to_csv, "a") as f:
                self.tables[path_to_csv.name].to_csv(f, header=isfirst, index=False)

        append_csv(self.path / self.OUTPUT_FILE_NAME, self.pending_output)
        append_csv(self.path / self.INPUT_FILE_NAME, self.pending_input)


class MungeToHDF(Device):
    """This is where ugly but necessary sausage making organizes
    in a file output with a key in the root database.

    The following conversions to relational files are attempted in order to
    convert each value in the row dictionary:

    1. Text containing a valid file or directory *outside* of the root data
       directory is made relational by moving the file or directory into
       the current row. The text is replaced with the updated relative path;
    2. Text longer than `text_relational_min` is dumped into a relational
       text file;
    3. 1- or 2-D data is converted to a pandas Series or DataFrame, and
       dumped into a relational file defined by the extension set by
       `nonscalar_file_type`

    """

    resource = value.Path(help="hdf file location")
    key_fmt = value.str(
        "{id} {host_time}",
        help="format for linked data in the root database (keyed on column)",
    )

    def open(self):
        import h5py

        self.backend = h5 = h5py.File(self.resource, "a")

    def close(self):
        self.backend.close()

    def __call__(self, index, row):
        """
        Break special cases of row items that need to be stored in
        relational files. The row valueis replaced in-place with the relative
        path to the data saved on disk.

        Arguments:
            row: dictionary of {'entry_name': "entry_value"} pairs
        Returns:
            the row dictionary, replacing special entries with the relative path to the saved data file
        """

        def is_path(v):
            if not isinstance(v, str):
                return False
            try:
                return os.path.exists(v)
            except ValueError:
                return False

        for name, value in row.items():
            # A path outside the relational database tree
            if is_path(value):
                # A file or directory that should be moved in
                row[name] = self._from_external_file(name, value, index, row)

            # A long string that should be written to a text file
            elif isinstance(value, (str, bytes, Number)):
                row[name] = value

            elif hasattr(value, "__len__") or hasattr(value, "__iter__"):
                # vector, table, matrix, etc.
                row[name] = self._from_nonscalar(name, value, index, row)

            else:
                self._logger.warning(
                    fr"unrecognized type for row entry '{name}' with type {repr(value)}"
                )
                row[name] = value

        return row

    def save_metadata(self, name, key_func, **extra):
        def process_value(value, key_name):
            if isinstance(value, (str, bytes)):
                return value
            elif hasattr(value, "__len__") or hasattr(value, "__iter__"):
                if not hasattr(value, "__len__") or len(value) > 0:
                    self._from_nonscalar(key_name, value)
                else:
                    return ""
            else:
                return value

        summary = dict(extra)
        for owner, owner_name in name.items():
            if owner_name.endswith("_values"):
                for trait in owner:
                    summary[key_func(owner_name, trait.name)] = getattr(
                        owner, trait.name
                    )
        summary = {k: process_value(v, k) for k, v in summary.items()}

        metadata = pd.DataFrame([summary], index=["Value"]).T
        metadata.astype(str).to_hdf(self.resource, key="metadata", append=True)

    def _from_nonscalar(self, name, value, index=0, row=None):
        """Write nonscalar (potentially array-like, or a python object) data
        to a file, and return a path to the file

        Arguments:
            name: name of the entry to write, used as the filename
            value: the object containing array-like data
            row: row dictionary, or None (the default) to write to the metadata folder
        Returns:
            the path to the file, relative to the directory that contains the root database
        """

        key = self._get_key(name, index, row)

        try:
            df = pd.DataFrame(value)
            if df.shape[0] == 0:
                df = pd.DataFrame([df])
        except BaseException:
            # We couldn't make a DataFrame
            self._logger.error(
                f"Failed to form DataFrame from {repr(name)}; pickling object instead"
            )
            self.backend[key] = pickle.dumps(value)

        else:
            df.to_hdf(self.resource, key=key)

        return key

    def _from_external_file(self, name, old_path, index=0, row=None, ntries=10):

        with open(old_path, "rb") as f:
            return f.read()

    def _get_key(self, name, index, row):
        if row is None:
            return f"/metadata {name}"
        else:
            return "/" + self.key_fmt.format(id=index, **row) + " " + name


class HDFLogger(RelationalTableLogger):
    """Store data and activity from value and property sets and gets to disk
    into a root database formatted as an HDF file.

    This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path (str): Base path to use for the root database
        append (bool): Whether to append to the root database if it already exists (otherwise, raise IOError)
        key_fmt (str): format to use for keys in the h5

    """

    KEY_OUTPUT = 'output'
    KEY_INPUT = 'input'

    nonscalar_file_type = "csv"

    def __init__(
        self,
        path,
        *,
        append=False,
        key_fmt="{id} {host_time}",
        git_commit_in=None,
        # **metadata
    ):
        if str(path).endswith(".h5"):
            path = Path(path)
        else:
            path = Path(str(path) + ".h5")

        super().__init__(
            path=path, append=append, git_commit_in=git_commit_in,
        )

        # Switch to the HDF munger
        self.munge = MungeToHDF(path, key_fmt=key_fmt)

    def open(self):
        """Instead of calling `open` directly, consider using
        `with` statements to guarantee proper disconnection
        if there is an error. For example, the following
        sets up a connected instance::

            with HDFLogger('my.csv') as db:
                ### do the data acquisition here
                pass

        would instantiate a `CSVLogger` instance, and also guarantee
        a final attempt to write unwritten data is written, and that
        the file is closed when exiting the `with` block, even if there
        is an exception.
        """
        self.df = None

    def close(self):
        self._write_root()

    def _write_root(self):
        """Write queued rows of data to csv. This is called automatically on :func:`close`, or when
        exiting a `with` block.

        If the class was created with overwrite=True, then the first call to _write_root() will overwrite
        the preexisting file; subsequent calls append.
        """

        def write_table(key, data):
            print('write table! ', key, data)
            if len(data) == 0:
                return
            isfirst = self.df is None
            pending = pd.DataFrame(data)
            pending.index.name = self.index_label
            pending.index += self.output_index
            if isfirst:
                print('first!')
                self.df = pending
            else:
                print(self.df.columns, pending.columns)
                self.df = self.df.append(pending).loc[self.output_index :]
            self.df.sort_index(inplace=True)
            self.output_index = self.df.index[-1]

            self.df.to_hdf(self.path, key=key, append=self._append)

        write_table(self.KEY_OUTPUT, self.pending_output)
        write_table(self.KEY_INPUT, self.pending_input)


class SQLiteLogger(RelationalTableLogger):
    """Store data and property traits to disk into an an sqlite database.

    This extends :class:`Aggregator` to support

    #. queuing aggregate property trait of devices by lists of dictionaries;
    #. custom metadata in each queued aggregate property trait entry; and
    #. custom response to non-scalar data (such as relational databasing).

    Arguments:
        path (str): Base path to use for the root database
        overwrite (bool): Whether to overwrite the root database if it exists (otherwise, append)
        text_relational_min: Text with at least this many characters is stored as a relational text file instead of directly in the database
        force_relational: A list of columns that should always be stored as relational data instead of directly in the database
        nonscalar_file_type: The data type to use in non-scalar (tabular, vector, etc.) relational data
        metadata_dirname: The name of the subdirectory that should be used to store metadata (device connection parameters, etc.)
        tar: Whether to store the relational data within directories in a tar file, instead of subdirectories
    """

    index_label = "id"  # Don't change this or sqlite breaks :(
    ROOT_FILE_NAME = "root.db"
    OUTPUT_TABLE_NAME = "output"
    _columns = None
    inprogress = {}
    committed = {}
    output_index = 0
    _engine = None

    def open(self):
        """Instead of calling `open` directly, consider using
        `with` statements to guarantee proper disconnection
        if there is an error. For example, the following
        sets up a connected instance::

            with SQLiteLogger('my.db') as db:
                ### do the data acquisition here
                pass

        would instantiate a `SQLiteLogger` instance, and also guarantee
        a final attempt to write unwritten data is written, and that
        the file is closed when exiting the `with` block, even if there
        is an exception.
        """

        if self.path.exists():
            root = str(self.path.absolute())
            if not self.path.is_dir():
                raise IOError(
                    f"the root data directory path '{root}' already exists, and is not a directory."
                )

            try:
                # test writes
                with open(self.path / "_", "wb"):
                    pass
                (self.path / "_").unlink()
                ex = None
            except IOError as e:
                ex = e
            if ex:
                raise ex

        #        if not self.path.lower().endswith('.db'):
        #            self.path += '.db'
        os.makedirs(self.path, exist_ok=True)
        path = os.path.join(self.path, self.ROOT_FILE_NAME)

        # Create an engine via sqlalchemy
        from sqlalchemy import create_engine  # database connection

        self._engine = create_engine("sqlite:///{}".format(path))

        if os.path.exists(path):
            if self._append:
                # read the last row to 1) list the columns and 2) get index
                query = f"select * from {self.OUTPUT_TABLE_NAME} order by {self.index_label} desc limit 1"

                df = pd.read_sql_query(query, self._engine, index_col=self.index_label)
                self._columns = df.columns
                self.output_index = df.index[-1] + 1
            else:
                raise IOError(
                    f"root table already exists at '{path}', but append=False"
                )
        else:
            self._columns = None
        self.inprogress = {}
        self.committed = {}

    def close(self):
        try:
            self._write_root()
        finally:
            self._engine.dispose()

    def _write_root(self):
        """Write queued rows of data to the database. This also is called automatically on :func:`close`, or when
        exiting a `with` block.

        If the class was created with overwrite=True, then the first call to _write_root() will overwrite
        the preexisting file; subsequent calls append.
        """

        if len(self.pending_output) == 0:
            return

        # Put together a DataFrame from self.pending that is guaranteed
        # to include the columns defined in self._columns
        traits = pd.DataFrame(self.pending_output)
        traits.index = traits.index + self.output_index
        blank = pd.DataFrame(columns=self._columns)

        # Check for new columns, and insert into the database
        if self._columns is None:
            new_columns = []
        else:
            trait_cols = [c.lower() for c in traits.columns]
            blank_cols = [c.lower() for c in blank.columns]
            new_columns = list(set(trait_cols).difference(blank_cols))

        for c in new_columns:
            self._logger.debug(f"inserting new column '{c}'")
            column_type = self._sql_type_name(traits[c])
            with self._engine.connect() as con:
                query = f"alter table {self.OUTPUT_TABLE_NAME} add column {c} {column_type} default NULL"
                con.execute(query)

        # Form the new database row
        df = blank.append(traits, sort=True)
        df.sort_index(inplace=True)
        self._columns = df.columns

        from sqlalchemy.exc import ArgumentError

        # Append to db
        try:
            df.to_sql(
                self.OUTPUT_TABLE_NAME,
                self._engine,
                if_exists="append",
                index=True,
                index_label=self.index_label,
            )
        except ArgumentError:
            self._logger.error(f"failed to convert index label {self.index_label}")
            raise ArgumentError
        except BaseException as e:
            raise e

        self.output_index += 1

    def key(self, name, attr):
        """The key determines the SQL column name. df.to_sql does not seem
        to support column names that include spaces
        """
        return f"{name.replace(' ', '_')}_{attr}"

    def _sql_type_name(self, col):
        from pandas.io.sql import _SQL_TYPES

        col_type = self._get_notnull_col_dtype(col)
        if col_type == "timedelta64":
            warnings.warn(
                "the 'timedelta' type is not supported, and will be \
                           written as integer values (ns frequency) to the\
                           database."
            )
            col_type = "integer"

        elif col_type == "datetime64":
            col_type = "datetime"

        elif col_type == "empty":
            col_type = "string"

        elif col_type == "complex":
            raise ValueError("Complex datatypes not supported")

        if col_type not in _SQL_TYPES:
            col_type = "string"

        return _SQL_TYPES[col_type]

    def _get_notnull_col_dtype(self, col):
        """
        Infer datatype of the Series col.  In case the dtype of col is 'object'
        and it contains NA values, this infers the datatype of the not-NA
        values.  Needed for inserting typed data containing NULLs, GH8778.
        """

        col_for_inference = col
        if col.dtype == "object":
            notnulldata = col[~pd.isnull(col)]
            if len(notnulldata):
                col_for_inference = notnulldata

        return pd.api.types.infer_dtype(col_for_inference)


def to_feather(data, path):
    """
    Write a dataframe to a feather file on disk. Any index will be moved
    to a column, index and column name metadata will be removed,
    and columns names will be changed to a string.

    Arguments:
        data: dataframe to write to disk
        path: path to file to write
    Returns:
        None

    """
    import numpy as np

    iname, data.index.name = data.index.name, None
    cname, data.columns.name = data.columns.name, None
    try:
        if not (
            data.index.is_monotonic
            and data.index[0] == 0
            and data.index[-1] == data.shape[0] - 1
        ):
            data = data.reset_index()
        data.columns = np.array(data.columns).astype(np.str)
        data.to_feather(path)
    finally:
        data.index.name = iname
        data.columns.name = cname


def read_sqlite(
    path,
    table_name=SQLiteLogger.OUTPUT_TABLE_NAME,
    columns=None,
    nrows=None,
    index_col=RelationalTableLogger.index_label,
):
    """Wrapper to that uses pandas.read_sql_table to load a table from an sqlite database at the specified path.

    Arguments:
        path: sqlite database path
        table_name: name of table in the sqlite database
        columns: columns to query and return, or None (default) to return all columns
        nrows: number of rows of data to read, or None (default) to return all rows
        index_col: the name of the column to use as the index
    Returns:
        pandas.DataFrame instance containing data loaded from `path`
    """

    from sqlalchemy import create_engine

    engine = create_engine(f"sqlite:///{path}")
    df = pd.read_sql_table(table_name, engine, index_col=index_col, columns=columns)
    engine.dispose()
    if nrows is not None:
        df = df.iloc[:nrows]
    return df


def read(path_or_buf, columns=None, nrows=None, format="auto", **kws):
    """Read tabular data from a file in one of various formats
    using pandas.

    Arguments:
        path (str): path to the  data file.
        columns: a column or iterable of multiple columns to return from the data file, or None (the default) to return all columns
        nrows: number of rows to read at the beginning of the table, or None (the default) to read all rows
        format (str): data file format, one of ['pickle','feather','csv','json','csv'], or 'auto' (the default) to guess from the file extension
        kws: additional keyword arguments to pass to the pandas read_<ext> function matching the file extension
    Returns:
        pandas.DataFrame instance containing data read from file
    """

    from pyarrow.feather import read_feather

    reader_guess = {
        "p": pd.read_pickle,
        "pickle": pd.read_pickle,
        "db": read_sqlite,
        "sqlite": read_sqlite,
        "json": pd.read_json,
        "csv": pd.read_csv,
    }

    try:
        reader_guess.update({"f": read_feather, "feather": read_feather})
    except BaseException:
        warnings.warn(
            "feather format is not available in this pandas installation, and will not be supported in labbench"
        )

    if isinstance(path_or_buf, (str, Path)):
        path_or_buf = str(path_or_buf)
        if os.path.getsize(path_or_buf) == 0:
            raise IOError("file is empty")

        if format == "auto":
            format = os.path.splitext(path_or_buf)[-1][1:]
    else:
        if format == "auto":
            raise ValueError(
                "can only guess format for string path - specify extension"
            )

    try:
        reader = reader_guess[format]
    except KeyError as e:
        raise Exception(f"couldn't guess a reader from extension of file {path_or_buf}")

    if reader == read_sqlite:
        return reader(path_or_buf, columns=columns, nrows=nrows, **kws)
    elif reader == pd.read_csv:
        return reader(path_or_buf, usecols=columns, nrows=nrows, **kws)
    else:
        if columns is None:
            columns = slice(None, None)
        return reader(path_or_buf, **kws)[columns].iloc[:nrows]


class MungeTarReader:
    tarnames = "data.tar", "data.tar.gz", "data.tar.bz2", "data.tar.lz4"

    def __init__(self, path, tarname="data.tar"):
        import tarfile

        self.tarfile = tarfile.open(os.path.join(path, tarname), "r")

    def __call__(self, key, *args, **kws):
        key = key.replace("\\\\", "\\")

        for k in key, key.replace("\\", "/").replace("//", "/"):
            try:

                ext = os.path.splitext(key)[1][1:]
                return read(self.tarfile.extractfile(k), format=ext, *args, **kws)
            except KeyError as e:
                ex = e
                continue
            else:
                break
        else:
            raise ex


class MungeDirectoryReader:
    def __init__(self, path):
        self.path = Path(path)

    def __call__(self, key, *args, **kws):
        return read(os.path.join(self.path, key), *args, **kws)


class MungeReader:
    """Guess the type of munging performed on the relational data, and return
    a reader suited to loading that file.

    """

    """ TODO: Make this smarter, perhaps by trying to read an entry from
        the root database
    """

    def __new__(cls, path):
        dirname = os.path.dirname(path)

        for n in MungeTarReader.tarnames:
            if os.path.exists(os.path.join(dirname, n)):
                return MungeTarReader(dirname, tarname=n)
        return MungeDirectoryReader(dirname)


def read_relational(
    path,
    expand_col,
    root_cols=None,
    target_cols=None,
    root_nrows=None,
    root_format="auto",
    prepend_column_name=True,
):
    """Flatten a relational database table by loading the table located each row of
    `root[expand_col]`. The value of each column in this row
    is copied to the loaded table. The columns in the resulting table generated
    on each row are downselected
    according to `root_cols` and `target_cols`. Each of the resulting tables
    is concatenated and returned.

    The expanded dataframe may be very large, making downselecting a practical
    necessity in some scenarios.

    TODO: Support for a list of expand_col?

    :param pandas.DataFrame root: the root database, consisting of columns containing data and columns containing paths to data files
        expand_col (str): the column in the root database containing paths to    data files that should be expanded
        root_cols: a column (or array-like iterable of multiple columns) listing the root columns to include in the expanded dataframe, or None (the default) pass all columns from `root`
        target_cols: a column (or array-like iterable of multiple columns) listing the root columns to include in the expanded dataframe, or None (the default) to pass all columns loaded from each root[expand_col]
        root_path: a string containing the full path to the root database (to help find the relational files)
        prepend_column_name (bool): whether to prepend the name of the expanded column from the root database
    Returns:
        the expanded dataframe

    """

    # if not isinstance(root, (pd.DataFrame,pd.Series)):
    #     raise ValueError('expected root to be a DataFrame instance, but it is {} instead'\
    #                      .format(repr(type(root))))
    if not isinstance(expand_col, str):
        raise ValueError(f"expand_col must a str, not {type(expand_col)}")

    if root_cols is not None:
        root_cols = list(root_cols) + [expand_col]
    root = read(path, columns=root_cols, nrows=root_nrows, format=root_format)
    reader = MungeReader(path)

    def generate():
        for i, row in root.iterrows():
            if row[expand_col] is None or len(row[expand_col]) == 0:
                continue

            sub = reader(row[expand_col], columns=target_cols)

            if prepend_column_name:
                prepend = expand_col + "_"
            else:
                prepend = ""

            # Rename columns
            sub.columns = [prepend + c for c in sub.columns]
            sub[prepend + "id"] = sub.index

            # # Downselect root columns
            # if root_cols is not None:
            #     row = row[root_cols]

            # Add in columns from the root
            for c, v in row.iteritems():
                sub[c] = v

            yield sub

    return pd.concat(generate(), ignore_index=True, sort=True)



# -*- coding: UTF-8 -*-

# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.

from labbench.util import sequentially
import sys

if ".." not in sys.path:
    sys.path.insert(0, "..")

import unittest
import importlib

import time

try:
    from .emulate import EmulatedVISADevice
except:
    from emulate import EmulatedVISADevice
from contextlib import contextmanager
import labbench as lb
import numpy as np


class LaggyInstrument(EmulatedVISADevice):
    """ A fake "instrument" with traits and fetch methods
    """

    # Connection and driver value traits
    delay = lb.value.float(0, min=0, help="connection time")
    fetch_time = lb.value.float(0, min=0, help="fetch time")
    fail_disconnect = lb.value.bool(
        False, help="True to raise DivideByZero on disconnect"
    )

    variable_setting = lb.value.int(0)

    def open(self):
        self.perf = {}
        t0 = time.perf_counter()
        lb.sleep(self.delay)
        self.perf["open"] = time.perf_counter() - t0

    def fetch(self):
        """ Return the argument after a 1s delay
        """
        import pandas as pd

        lb.logger.info(f"{self}.fetch start")
        t0 = time.perf_counter()
        lb.sleep(self.fetch_time)
        self.perf["fetch"] = time.perf_counter() - t0
        return pd.Series([1, 2, 3, 4, 5, 6])

    def dict(self):
        return {self.resource: self.resource}

    def none(self):
        """ Return None
        """
        return None

    def close(self):
        if self.fail_disconnect:
            1 // 0


class Rack1(lb.Rack):
    dev1: LaggyInstrument = LaggyInstrument()
    dev2: LaggyInstrument = LaggyInstrument()

    def setup(self, param1: float):
        pass

    def arm(self):
        self.dev1.variable_setting = self.dev1.variable_setting + 2

    def open(self):
        time.sleep(0.25)


rack = Rack1(dev1=LaggyInstrument(resource="address"))


class Rack2(lb.Rack):
    dev: LaggyInstrument = LaggyInstrument()

    def setup(self):
        return "rack 2 - setup"
        return self.dev.dict()

    def acquire(self, *, param1):
        return "rack 3 - acquire"

    def fetch(self, *, param2: int = 7):
        return self.dev.fetch()

    def open(self):
        time.sleep(0.25)


rack = Rack1()

with rack:
    #
    pass


class Rack3(lb.Rack):
    # this is unset (no =), so it _must_ be passed as an argument to instantiate, i.e.,
    #
    # >>> Rack3(dev=MyDevice)
    dev: LaggyInstrument

    # notice this operation requires param2 and param3
    def acquire(self, *, param3, param2=7):
        pass

    def fetch(self, *, param4):
        return self.dev.fetch()

    def open(self):
        time.sleep(0.25)


class MyRack(lb.Rack):
    # config = lb.Configuration('test-config')

    # db = lb.HDFLogger(
    #     path=time.strftime(f"%Y-%m-%d_%Hh%Mm%Ss"),  # Path to new directory that will contain containing all files
    #     append=True,  # `True` --- allow appends to an existing database; `False` --- append
    #     key_fmt='{id} {host_time}',  # Format string that generates relational data (keyed on data column)
    # )

    db = lb.CSVLogger(
        path=time.strftime(
            f"test db/test-rack %Y-%m-%d_%Hh%Mm%Ss"
        ),  # Path to new directory that will contain containing all files
        append=True,  # `True` --- allow appends to an existing database; `False` --- append
        text_relational_min=1024,  # Minimum text string length that triggers relational storage
        force_relational=[
            "host_log"
        ],  # Data in these columns will always be relational
        dirname_fmt="{id} {host_time}",  # Format string that generates relational data (keyed on data column)
        nonscalar_file_type="csv",  # Default format of numerical data, when possible
        metadata_dirname="metadata",  # metadata will be stored in this subdirectory
        tar=False,  # `True` to embed relational data folders within `data.tar`
    )

    # Devices
    inst1: LaggyInstrument = LaggyInstrument(resource="a", delay=0.12)
    inst2: LaggyInstrument = LaggyInstrument(resource="b", delay=0.06)

    # Test procedures
    rack1 = Rack1(dev1=inst1, dev2=inst2)
    rack2 = Rack2(dev=inst1)
    rack3 = Rack3(dev=inst2)

    run = lb.Sequence(
        setup=(rack1.setup, rack2.setup),  # 2 methods run concurrently
        arm=rack1.arm,
        acquire1=rack2.acquire,
        acquire2=rack3.acquire,
        fetch=(rack2.fetch, rack3.fetch),  # 2 also run concurrently
        finish=db.new_row,
    )

    def open(self):
        pass

    def close(self):
        pass


with rack:
    pass


import inspect

if __name__ == "__main__":
    lb.show_messages("debug")
    lb.util._force_full_traceback(True)

    # # Testbed = lb.Rack.take_module('module_as_testbed')
    Testbed = MyRack

    # lst = CommentedSeq(['a', 'b'])
    # lst.yaml_add_eol_comment("foo", 0, 0)
    # lst.yaml_add_eol_comment("bar\n\n", 1)
    # data["list_of_elements_side_comment"] = lst
    # data.yaml_set_comment_before_after_key("list_of_elements_side_comment", "\n")

    # lst = CS(['a', 'b'])
    # lst.yaml_set_comment_before_after_key(0, "comment 1", 2)
    # lst.yaml_set_comment_before_after_key(1, "comment 2", 2)
    # data["list_of_elements_top_comment"] = lst

    # Testbed.config.make_templates()

    # # Testbed.config._rack_defaults(Testbed)

    # with lb.stopwatch('test connection'):
    #     with Testbed() as testbed:
    #         # with MyRack() as testbed:
    #         testbed.inst2.delay = 0.07
    #         testbed.inst1.delay = 0.12

    #         testbed.run.iterate_from_csv('setup/MyRack.run.csv')

    #         # for i in range(3):
    #         #     # Run the experiment
    #         #     ret = testbed.run(rack1_param1=1, rack2_param1=2, rack3_param2=3,
    #         #                       rack3_param3=4, rack2_param2=5, rack3_param4=6)
    #         #
    #         #     testbed.db()

    # df = lb.read(testbed.db.path/'master.db')
    # df.to_csv(testbed.db.path/'master.csv')

    # config_root = Path('test-config')
    # config_root.mkdir(parents=True, exist_ok=True)

    # # default devices
    # defaults = {
    #     dev_name: {k:getattr(dev, k) for k in dev._value_attrs}
    #     for dev_name, dev in MyRack._devices.items()
    # }
    # with open(config_root/'devices.yaml', 'w') as f:
    #     yaml.dump(defaults, f)

    # defaults, annots, methods, names = MyRack.config.parameters(MyRack)

    # with open(config_root/'rack arguments.yaml', 'w') as f:
    #     f.write('## Comment or remove lines to add as columns in sequence tables\n\n')
    #     for k, default in defaults.items():
    #         if default is lb._rack.EMPTY:
    #             s = f'# {k}: \n'
    #             f.write(s)
    #         else:
    #             s = yaml.dump({k: default}, f)
    #         if annots[k] is not lb._rack.EMPTY:
    #             before, *after = s.split(b'\n', 1)
    #             s = b'\n'.join([before+f" # {annots[k].__qualname__}"] + after)



import importlib
import sys
import time

if ".." not in sys.path:
    sys.path.insert(0, "..")
import labbench as lb

lb = importlib.reload(lb)

from emulate import EmulatedVISADevice


class LaggyInstrument(EmulatedVISADevice):
    """ A mock "instrument"
    with value and property traits to
    demonstrate the process of value trait
    up a measurement.
    """

    # Connection and driver value traits
    delay = lb.value.float(default=0, min=0, help="connection time (s)")
    fetch_time = lb.value.float(default=0, min=0, help="fetch time (s)")
    fail_disconnect = lb.value.bool(
        default=False, help="raise DivideByZero on disconnect?"
    )

    def open(self):
        self.perf = {}
        t0 = time.perf_counter()
        lb.sleep(self.delay)
        self.perf["open"] = time.perf_counter() - t0

    def fetch(self):
        """ Return the argument after a 1s delay
        """
        lb.logger.info(f"{self}.fetch start")
        t0 = time.perf_counter()
        lb.sleep(self.fetch_time)
        self.perf["fetch"] = time.perf_counter() - t0
        return self.fetch_time

    def dict(self):
        return {self.resource: self.resource}

    def none(self):
        """ Return None
        """
        return None

    def close(self):
        if self.fail_disconnect:
            1 / 0


class Rack1(lb.Rack):
    dev1: lb.Device = LaggyInstrument()
    dev2: lb.Device = LaggyInstrument()

    def setup(self, param1):
        pass

    def arm(self):
        pass


class Rack2(lb.Rack):
    dev: lb.Device = LaggyInstrument()

    def setup(self):
        return "rack 2 - setup"
        return self.dev.dict()

    def acquire(self, *, param1):
        return "rack 3 - acquire"

    def fetch(self, *, param2=7):
        return self.dev.fetch()


class Rack3(lb.Rack):
    dev: lb.Device = LaggyInstrument()

    def acquire(self, *, param2=7, param3):
        pass

    def fetch(self, *, param4):
        self.dev.fetch()


db: lb._data.RelationalTableLogger = lb.SQLiteLogger(
    "data",  # Path to new directory that will contain containing all files
    append=True,  # `True` --- allow appends to an existing database; `False` --- append
    text_relational_min=1024,  # Minimum text string length that triggers relational storage
    force_relational=["host_log"],  # Data in these columns will always be relational
    dirname_fmt="{id} {host_time}",  # Format string that generates relational data (keyed on data column)
    nonscalar_file_type="csv",  # Default format of numerical data, when possible
    metadata_dirname="metadata",  # metadata will be stored in this subdirectory
    tar=False,  # `True` to embed relational data folders within `data.tar`
)

# Devices
inst1: LaggyInstrument = LaggyInstrument(resource="a", delay=0.12)
inst2: LaggyInstrument = LaggyInstrument(resource="b", delay=0.06)

# Test procedures
rack1 = Rack1(dev1=inst1, dev2=inst2)
rack2 = Rack2(dev=inst1)
rack3 = Rack3(dev=inst2)

run = lb.Sequence(
    setup=(rack1.setup, rack2.setup),  # executes these 2 methods concurrently
    arm=(rack1.arm),
    acquire1=rack2.acquire,
    acquire2=rack3.acquire,  # executes these 2 sequentially
    fetch=(rack2.fetch, rack3.fetch),
    finish=(db.new_row),
)



import contextlib
from . import util as util, value as value
from ._device import Device as Device
from ._traits import observe as observe
from _typeshed import Incomplete
from collections.abc import Generator
win32com: Incomplete
pyvisa: Incomplete


class ShellBackend(Device):

    def __init__(self, resource: str='str', binary_path: str='NoneType', timeout: str='int'):
        ...
    binary_path: Incomplete
    timeout: Incomplete

    @classmethod
    def __imports__(cls) -> None:
        ...
    backend: Incomplete

    def open(self) -> None:
        ...

    def run(
        self,
        *argv,
        pipe: bool=...,
        background: bool=...,
        check_return: bool=...,
        check_stderr: bool=...,
        respawn: bool=...,
        timeout: Incomplete | None=...
    ):
        ...

    def read_stdout(self, wait_for: int=...):
        ...

    def write_stdin(self, text) -> None:
        ...

    def kill(self) -> None:
        ...

    def running(self):
        ...

    def clear_stdout(self) -> None:
        ...

    def close(self) -> None:
        ...


class DotNetDevice(Device):

    def __init__(self, resource: str='str'):
        ...
    library: Incomplete
    dll_name: Incomplete
    dll: Incomplete

    def open(self) -> None:
        ...

    def open(self) -> None:
        ...


class LabviewSocketInterface(Device):

    def __init__(
        self,
        resource: str='str',
        tx_port: str='int',
        rx_port: str='int',
        delay: str='int',
        timeout: str='int',
        rx_buffer_size: str='int'
    ):
        ...
    resource: Incomplete
    tx_port: Incomplete
    rx_port: Incomplete
    delay: Incomplete
    timeout: Incomplete
    rx_buffer_size: Incomplete
    backend: Incomplete

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    def write(self, msg) -> None:
        ...

    def set_key(self, key, value, name) -> None:
        ...

    def read(self, convert_func: Incomplete | None=...):
        ...

    def clear(self) -> None:
        ...


class SerialDevice(Device):

    def __init__(
        self,
        resource: str='str',
        timeout: str='int',
        write_termination: str='bytes',
        baud_rate: str='int',
        parity: str='bytes',
        stopbits: str='int',
        xonxoff: str='bool',
        rtscts: str='bool',
        dsrdtr: str='bool'
    ):
        ...
    resource: Incomplete
    timeout: Incomplete
    write_termination: Incomplete
    baud_rate: int
    parity: Incomplete
    stopbits: Incomplete
    xonxoff: Incomplete
    rtscts: Incomplete
    dsrdtr: Incomplete
    backend: Incomplete

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    @classmethod
    def from_hwid(cls, hwid: Incomplete | None=..., *args, **connection_params):
        ...

    @staticmethod
    def list_ports(hwid: Incomplete | None=...):
        ...


class SerialLoggingDevice(SerialDevice):

    def __init__(
        self,
        resource: str='str',
        timeout: str='int',
        write_termination: str='bytes',
        baud_rate: str='int',
        parity: str='bytes',
        stopbits: str='int',
        xonxoff: str='bool',
        rtscts: str='bool',
        dsrdtr: str='bool',
        poll_rate: str='float',
        data_format: str='bytes',
        stop_timeout: str='float',
        max_queue_size: str='int'
    ):
        ...
    poll_rate: Incomplete
    data_format: Incomplete
    stop_timeout: Incomplete
    max_queue_size: Incomplete

    def configure(self) -> None:
        ...

    def start(self) -> None:
        ...

    def stop(self) -> None:
        ...

    def running(self):
        ...

    def fetch(self):
        ...

    def clear(self) -> None:
        ...

    def close(self) -> None:
        ...


class TelnetDevice(Device):

    def __init__(self, resource: str='str', timeout: str='int'):
        ...
    resource: Incomplete
    timeout: Incomplete

    @classmethod
    def __imports__(cls) -> None:
        ...
    backend: Incomplete

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...


class VISADevice(Device):

    def __init__(
        self,
        resource: str='str',
        read_termination: str='str',
        write_termination: str='str'
    ):
        ...
    read_termination: Incomplete
    write_termination: Incomplete
    identity: Incomplete
    options: Incomplete

    def status_byte(self):
        ...

    @classmethod
    def __imports__(cls) -> None:
        ...
    backend: Incomplete

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    @classmethod
    def list_resources(cls):
        ...

    def write(self, msg: str):
        ...

    def query(self, msg: str, timeout: Incomplete | None=...) -> str:
        ...

    def query_ascii_values(
        self,
        msg: str,
        type_,
        separator: str=...,
        container=...,
        delay: Incomplete | None=...,
        timeout: Incomplete | None=...
    ):
        ...

    def get_key(self, scpi_key, name: Incomplete | None=...):
        ...

    def set_key(self, scpi_key, value, name: Incomplete | None=...) -> None:
        ...

    def wait(self) -> None:
        ...

    def preset(self) -> None:
        ...

    def overlap_and_block(self, timeout: Incomplete | None=..., quiet: bool=...) -> Generator[
        None,
        None,
        None,
    ]:
        ...


    class suppress_timeout(contextlib.suppress):

        def __exit__(self, exctype, excinst, exctb):
            ...


class SimulatedVISADevice(VISADevice):

    def __init__(
        self,
        resource: str='str',
        read_termination: str='str',
        write_termination: str='str'
    ):
        ...
    yaml_source: Incomplete


class Win32ComDevice(Device):

    def __init__(self, resource: str='str'):
        ...
    com_object: Incomplete

    @classmethod
    def __imports__(cls) -> None:
        ...
    backend: Incomplete

    def open(self):
        ...



# This software was developed by employees of the National Institute of
# Standards and Technology (NIST), an agency of the Federal Government.
# Pursuant to title 17 United States Code Section 105, works of NIST employees
# are not subject to copyright protection in the United States and are
# considered to be in the public domain. Permission to freely use, copy,
# modify, and distribute this software and its documentation without fee is
# hereby granted, provided that this notice and disclaimer of warranty appears
# in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY OF ANY KIND, EITHER
# EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY
# THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND FREEDOM FROM
# INFRINGEMENT, AND ANY WARRANTY THAT THE DOCUMENTATION WILL CONFORM TO THE
# SOFTWARE, OR ANY WARRANTY THAT THE SOFTWARE WILL BE ERROR FREE. IN NO EVENT
# SHALL NIST BE LIABLE FOR ANY DAMAGES, INCLUDING, BUT NOT LIMITED TO, DIRECT,
# INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM,
# OR IN ANY WAY CONNECTED WITH THIS SOFTWARE, WHETHER OR NOT BASED UPON
# WARRANTY, CONTRACT, TORT, OR OTHERWISE, WHETHER OR NOT INJURY WAS SUSTAINED
# BY PERSONS OR PROPERTY OR OTHERWISE, AND WHETHER OR NOT LOSS WAS SUSTAINED
# FROM, OR AROSE OUT OF THE RESULTS OF, OR USE OF, THE SOFTWARE OR SERVICES
# PROVIDED HEREUNDER. Distributions of NIST software should also include
# copyright and licensing statements of any third-party software that are
# legally bundled with the code in compliance with the conditions of those
# licenses.
"""
This implementation is deeply intertwined obscure details of the python object
model. Consider starting with a close read of the documentation and exploring
the objects in an interpreter instead of reverse-engineering this code.
"""

from . import util

import typing
from warnings import warn
from functools import wraps
import validators as _val
from contextlib import contextmanager

from inspect import isclass
import inspect
import numbers
import re

# for common types
from pathlib import Path

Undefined = inspect.Parameter.empty

T = typing.TypeVar("T")


class ThisType(typing.Generic[T]):
    pass


class HasTraitsMeta(type):
    __cls_namespace__ = []

    @classmethod
    def __prepare__(cls, names, bases, **kws):
        """Prepare copies of cls._traits, to ensure that any traits defined
        in the definition don't clobber parents' traits.
        """
        ns = dict()
        if len(bases) >= 1:
            if hasattr(bases, "__children__"):
                ns["__children__"] = {}
            traits = {k: v.copy() for k, v in bases[0]._traits.items()}
            ns.update(traits)
            ns["_traits"] = traits
            HasTraitsMeta.__cls_namespace__.append(traits)
            return ns
        else:
            HasTraitsMeta.__cls_namespace__.append({})
            return dict(_traits=HasTraitsMeta.__cls_namespace__[-1])


class Trait:
    """base class for typed descriptors in Device classes. These
    implement type checking, casting, decorators, and callbacks.

    A Device instance supports two types of Traits:

    * A _value trait_ acts as an attribute variable in instantiated
      classes

    * A _property trait_ exposes set and get operations for a
      parameter in the API wrapped by the owning Device class.

    The trait behavior is determined by whether its owner is a Device or HasSettings
    instance.

    Arguments:
        default: the default value of the trait (value traits only)
        key: some types of Device take this input to determine automation behavior

        help: the Trait docstring
        label: a label for the quantity, such as units

    Arguments:
        sets: True if the trait supports writes
        gets: True if the trait supports reads

        cache: if True, interact with the device only once, then return copies (state traits only)
        only: value allowlist; others raise ValueError

    Arguments:
        allow_none: permit None values in addition to the specified type
        remap: a lookup table that maps the python type (keys) to a potentially different backend values (values) ,
                  in places of the to_pythonic and from_pythonic methods (property traits only)

    """

    ROLE_VALUE = "value"
    ROLE_PROPERTY = "property"
    ROLE_DATARETURN = "return"
    ROLE_UNSET = "unset"

    type = None
    role = ROLE_UNSET

    # keyword argument types and default values
    default: ThisType = Undefined
    key: Undefined = Undefined
    func: typing.Callable = None
    # role: str = ROLE_UNSET
    help: str = ""
    label: str = ""
    sets: bool = True
    gets: bool = True
    cache: bool = False
    only: tuple = tuple()
    allow_none: bool = False
    remap: dict = {}

    # If the trait is used for a state, it can operate as a decorator to
    # implement communication with a device
    _setter = None
    _getter = None
    _returner = None
    _decorated_funcs = []

    # __decorator_action__ = None

    def __init__(self, *args, **kws):
        if len(args) >= 1:
            if len(args) == 1 and self.role == self.ROLE_VALUE:
                if "default" in kws:
                    raise ValueError(f"duplicate 'default' argument in {self}")
                kws["default"] = args[0]
            elif len(args) == 1 and self.role == self.ROLE_DATARETURN:
                if "func" in kws:
                    raise ValueError("duplicate 'func' argument")
                kws["func"] = args[0]
            else:
                raise ValueError(f"no positional arguments supported")

        self.kws = dict(kws)
        self.metadata = {}
        self._decorated_funcs = []

        cls_defaults = {k: getattr(self, k) for k in self.__annotations__.keys()}

        if "default" in cls_defaults:
            cls_defaults["default"] = Undefined

        kws = dict(cls_defaults, **kws)

        # check role and related parameter dependencies
        if self.role == self.ROLE_VALUE:
            invalid_args = ("remap", "key", "func")
        elif self.role == self.ROLE_PROPERTY:
            invalid_args = ("default", "func")
        elif self.role == self.ROLE_DATARETURN:
            invalid_args = "default", "key", "sets", "gets"
        else:
            clsname = self.__class__.__qualname__
            raise ValueError(
                f"{clsname}.role must be one of {(self.ROLE_PROPERTY, self.ROLE_DATARETURN, self.ROLE_VALUE)}, not {repr(self.role)}"
            )

        for k in invalid_args:
            if k in cls_defaults and cls_defaults[k] != kws[k]:
                raise AttributeError(
                    f"keyword argument '{k}' is not allowed with {self.role}"
                )

        if self.role == self.ROLE_VALUE and kws["default"] is Undefined:
            # always go with None when this value is allowed, fallback to self.default
            kws["default"] = self.type()

        if self.role == self.ROLE_DATARETURN:
            if kws["func"] is not None:
                # apply a decorator
                self(kws["func"])

        if self.role in (self.ROLE_DATARETURN, self.ROLE_PROPERTY):
            # default Undefined so that cache will fill them in
            self.default = Undefined

        # Replace self.from_pythonic and self.to_pythonic with lookups in self.remap (if defined)
        try:
            if len(kws["remap"]) > 0:
                self.remap_inbound = {v: k for k, v in kws["remap"].items()}
            else:
                self.remap_inbound = {}
        except KeyError:
            raise

        if len(kws["remap"]) != len(self.remap_inbound):
            raise ValueError(f"'remap' has duplicate values")

        # set value traits
        for k, v in kws.items():
            setattr(self, k, v)

    @classmethod
    def __init_subclass__(cls, type=Undefined):
        """python triggers this call immediately after a Trait subclass
            is defined, allowing us to automatically customize its implementation.

        Arguments:
            type: the python type represented by the trait
        """
        if type is not Undefined:
            cls.type = type

        # complete the annotation dictionary with the parent
        annots = dict(
            getattr(cls.__mro__[1], "__annotations__", {}),
            **getattr(cls, "__annotations__", {}),
        )

        cls.__annotations__ = dict(annots)

        # # apply an explicit signature to cls.__init__
        # annots = {k: cls.type if v is ThisType else (k, v) \
        #           for k, v in annots.items()}

        # cls._arg_defaults = {k: getattr(cls, k)
        #                     for k in annots if hasattr(cls, k)}

        # if 'default' in cls._arg_defaults:
        #     cls._arg_defaults['default'] = Undefined

        # TODO: remove this
        # util.wrap_attribute(cls, '__init__', __init__, tuple(annots.keys()), cls._arg_defaults, 1, annots)

        # Help to reduce memory use by __slots__ definition (instead of __dict__)
        cls.__slots__ = [n for n in dir(cls) if not n.startswith("_")] + [
            "metadata",
            "kind",
            "name",
        ]

    def copy(self, new_type=None, **update_kws):
        if new_type is None:
            new_type = type(self)
        obj = new_type(**dict(self.kws, **update_kws))
        obj._getter = self._getter
        obj._setter = self._setter
        obj._returner = self._returner
        return obj

    ### Descriptor methods (called automatically by the owning class or instance)
    def __set_name__(self, owner_cls, name):
        """Immediately after an owner class is instantiated, it calls this
        method for each of its attributes that implements this method.

        Trait takes advantage of this to remember the owning class for debug
        messages and to register with the owner class.
        """
        # other owning objects may unintentionally become owners; this causes problems
        # if they do not implement the HasTraits object protocol
        if issubclass(owner_cls, HasTraits):
            # inspect module expects this name - don't play with it
            self.__objclass__ = owner_cls

            # Take the given name, unless we've bene tagged with a different
            self.name = name

            owner_cls._traits[name] = self

    def __init_owner_subclass__(self, owner_cls):
        """The owner calls this in each of its traits at the end of defining the subclass
        (near the end of __init_subclass__).
        has been called. Now it is time to ensure properties are compatible with the owner class.
        This is here --- not in __set_name__ --- because python
        obfuscates exceptions raised in __set_name__.

        This is also where we finalize selecting decorator behavior; is it a property or a method?
        """

        if self.role == self.ROLE_VALUE and len(self._decorated_funcs) > 0:
            raise AttributeError(
                f"tried to combine a default value and a decorator implementation in {self}"
            )
        elif self.role == self.ROLE_DATARETURN and len(self._decorated_funcs) == 0:
            raise AttributeError(f"decorate a method to tag its return data")
        elif len(self._decorated_funcs) == 0:
            return

        positional_argcounts = [
            f.__code__.co_argcount - len(f.__defaults__ or tuple())
            for f in self._decorated_funcs
        ]

        if self.role == self.ROLE_DATARETURN:
            for func, argcount in zip(self._decorated_funcs, positional_argcounts):
                if len(self.help.rstrip().strip()) == 0:
                    # take func docstring as default self.help
                    self.help = (func.__doc__ or "").rstrip().strip()

                self._returner = func

        elif self.role == self.ROLE_PROPERTY:
            if set(positional_argcounts) not in ({1}, {1, 2}, {2}):
                raise AttributeError(
                    f"a decorator implementation with @{self} must apply to a getter "
                    f"(above `def func(self)`) and/or setter (above `def func(self, value):`)"
                )
            for func, argcount in zip(self._decorated_funcs, positional_argcounts):
                doc = (func.__doc__ or "").strip().rstrip()
                if len(doc) > 0:
                    # take func docstring as default self.help
                    self.help = self.kws["help"] = doc

                if argcount == 1:
                    self._getter = func
                else:
                    self._setter = func

    def __init_owner_instance__(self, owner):
        # called by owner.__init__
        pass

    @util.hide_in_traceback
    def __set__(self, owner, value):
        # First, validate the pythonic types
        if not self.sets:
            raise AttributeError(f"{self.__str__()} cannot be set")

        # Validate the pythonic value
        if value is not None:
            # cast to self.type and validate
            value = Trait.to_pythonic(self, value)
            value = self.validate(value, owner)

            if len(self.only) > 0 and not self.contains(self.only, value):
                raise ValueError(
                    f"value '{value}' is not among the allowed values {repr(self.only)}"
                )
        elif self.allow_none:
            value = None
        else:
            raise ValueError(f"None value not allowed for trait '{repr(self)}'")

        try:
            value = self.from_pythonic(value)
        except BaseException as e:
            name = owner.__class__.__qualname__ + "." + self.name
            e.args = (e.args[0] + f" in attempt to set '{name}'",) + e.args[1:]
            raise e

        if self.role == self.ROLE_VALUE:
            # apply as a value trait
            owner.__set_value__(self.name, value)

        elif self.role == self.ROLE_PROPERTY:
            # convert to the outbound representation
            if len(self.remap) > 0:
                value = self.remap.get(value, value)

            # send to the device
            if self._setter is not None:
                # from the function decorated by this trait
                self._setter(owner, value)

            elif self.key is not None:
                # otherwise, use the owner's set_key
                owner.set_key(self.key, value, self.name)

            else:
                objname = owner.__class__.__qualname__ + "." + self.name
                raise AttributeError(
                    f"cannot set {objname}: no @{self.__repr__(owner_inst=owner)}."
                    f"setter and no key argument"
                )

        else:
            raise AttributeError(f"data return traits cannot be set")

        owner.__notify__(self.name, value, "set", cache=self.cache)

    @util.hide_in_traceback
    def __get__(self, owner, owner_cls=None):
        """Called by the class instance that owns this attribute to
        retreive its value. This, in turn, decides whether to call a wrapped
        decorator function or the owner's get_key method to retrieve
        the result.

        Returns:
            retreived value
        """

        # only continue to get the value if the __get__ was called for an owning
        # instance, and owning class is a match for what we were told in __set_name__.
        # otherwise, someone else is trying to access `self` and we
        # shouldn't get in their way.
        if owner is None or owner_cls.__dict__.get(
            self.name, None
        ) is not self.__objclass__.__dict__.get(self.name):
            # the __dict__ acrobatics avoids a recursive __get__ loop
            return self

        elif self.role == self.ROLE_DATARETURN:
            # inject the labbench Trait hooks into the return value
            @wraps(self._returner)
            def method(*args, **kws):
                value = self._returner(owner, *args, **kws)
                return self.__cast_get__(owner, value)

            return method

        elif not self.gets:
            # stop now if this is not a gets Trait
            raise AttributeError(f"{self.__repr__(owner_inst=owner)} is not gets")

        elif self.role == self.ROLE_VALUE:
            return owner.__get_value__(self.name)

        # from here on, operate as a property getter
        if self.cache and self.name in owner.__cache__:
            # return the cached value if applicable
            return owner.__cache__[self.name]

        elif self._getter is not None:
            # get value with the decorator implementation, if available
            value = self._getter(owner)

        else:
            # otherwise, get with owner.get_key, if available
            if self.key is None:
                # otherwise, 'get'
                objname = owner.__class__.__qualname__
                ownername = self.__repr__(owner_inst=owner)
                raise AttributeError(
                    f"to set the property {self.name}, decorate a method in {objname} or use the function key argument"
                )
            value = owner.get_key(self.key, self.name)

        # apply remapping as appropriate for the trait
        if len(self.remap_inbound) > 0:
            value = self.remap_inbound.get(value, value)

        return self.__cast_get__(owner, value, strict=False)

    @util.hide_in_traceback
    def __cast_get__(self, owner, value, strict=False):
        """Examine value and either return a valid pythonic value or raise an exception if it cannot be cast.

        Arguments:
            owner: the class that owns the trait
            value: the value we need to validate and notify
        :return:
        """
        if self.allow_none and value is None:
            pass
        else:
            # skip validation if None and None values are allowed
            try:
                value = self.to_pythonic(value)
            except BaseException as e:
                # name = owner.__class__.__qualname__ + '.' + self.name
                e.args = (
                    e.args[0]
                    + f" in attempt to get '{self.__repr__(owner_inst=owner)}'",
                ) + e.args[1:]
                raise e

            # Once we have a python value, give warnings (not errors) if the device value fails further validation
            if hasattr(owner, "_logger"):
                log = owner._logger.warning
            else:
                log = warn

            # TODO: This broke array-like data. Was it ever necessary?
            # if value != self.validate(value, owner):
            #     raise ValueError
            # except ValueError:
            #     log(f"'{self.__repr__(owner_inst=owner)}' {self.role} received the value {repr(value)}, " \
            #         f"which fails {repr(self)}.validate()")
            if value is None and not self.allow_none:
                log(
                    f"'{self.__repr__(owner_inst=owner)}' {self.role} received value None, which"
                    f"is not allowed for {repr(self)}"
                )
            if len(self.only) > 0 and not self.contains(self.only, value):
                log(
                    f"'{self.__repr__(owner_inst=owner)}' {self.role} received {repr(value)}, which"
                    f"is not in the valid value list {repr(self.only)}"
                )

        owner.__notify__(
            self.name, value, "get", cache=self.cache or (self.role == self.ROLE_VALUE)
        )

        return value

    @util.hide_in_traceback
    def to_pythonic(self, value):
        """Convert a value from an unknown type to self.type."""
        return self.type(value)

    @util.hide_in_traceback
    def from_pythonic(self, value):
        """convert from a python type representation to the format needed to communicate with the device"""
        return value

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        """This is the default validator, which requires that trait values have the same type as self.type.
            A ValueError is raised for other types.
            value: value to check
        Returns:
            a valid value
        """
        if not isinstance(value, self.type):
            typename = self.type.__qualname__
            valuetypename = type(value).__qualname__
            raise ValueError(
                f"{repr(self)} type must be '{typename}', not '{valuetypename}'"
            )
        return value

    def contains(self, iterable, value):
        return value in iterable

    ### Decorator methods
    @util.hide_in_traceback
    def __call__(self, func):
        """use the Trait as a decorator, which ties this Trait instance to evaluate a property or method in the
        owning class. you can specify
        """
        # only decorate functions.
        if not callable(func):
            raise Exception(
                f"object of type '{func.__class__.__qualname__}' must be callable"
            )

        self._decorated_funcs.append(func)

        # Register in the list of decorators, in case we are overwritten by an
        # overloading function
        if getattr(self, "name", None) is None:
            self.name = func.__name__
        if len(HasTraitsMeta.__cls_namespace__) > 0:
            HasTraitsMeta.__cls_namespace__[-1][func.__name__] = self

        # return self to ensure `self` is the value assigned in the class definition
        return self

    ### introspection
    ###
    def doc(self):
        params = self.doc_params(omit=["help", "default"])
        typename = "Any" if self.type is None else self.type.__qualname__

        doc = f"{self.name} ({typename}): {self.help}"
        if len(params) > 0:
            doc += f" ({params})"

        return doc

    def doc_params(self, omit=["help"]):
        pairs = []

        for name in self.__annotations__.keys():
            default = getattr(type(self), name)
            v = getattr(self, name)

            # skip uninformative debug info
            if name.startswith("_") or name in omit:
                continue

            # only show non-defaults
            v = getattr(self, name)
            if v == default:
                continue

            pairs.append(f"{name}={repr(v)}")

        return ",".join(pairs)

    def __repr__(self, omit=["help"], owner_inst=None):
        declaration = f"{self.role}.{type(self).__qualname__}({self.doc_params(omit)})"

        if owner_inst is None:
            return declaration
        else:
            return f"<{declaration} as {owner_inst}.{self.name}>"

    __str__ = __repr__

    def _owned_name(self, owner):
        if owner._owned_name is None:
            return type(owner).__qualname__ + "." + self.name
        else:
            return owner._owned_name + "." + self.name

    def update(self, obj=None, **attrs):
        """ returns `self` or (if `obj` is None) or `other`, after updating its keyword
        parameters with `attrs`
        """
        if obj is None:
            obj = self
        invalid_params = set(attrs).difference(obj.__dict__)
        if len(invalid_params) > 0:
            raise AttributeError(
                f"{obj} does not have the parameter(s) {invalid_params}"
            )

        obj.__dict__.update(attrs)
        return obj


Trait.__init_subclass__()


@contextmanager
def hold_trait_notifications(owner):
    def skip_notify(name, value, type, cache):
        old = owner.__cache__.setdefault(name, Undefined)

        msg = dict(new=value, old=old, owner=owner, name=name, type=type, cache=cache)

        owner.__cache__[name] = value

    original, owner.__notify__ = owner.__notify__, skip_notify
    yield
    owner.__notify__ = original


class HasTraits(metaclass=HasTraitsMeta):
    __notify_list__ = {}
    __cls_namespace__ = {}

    def __init__(self, **values):
        # who is informed on new get or set values
        self.__notify_list__ = {}

        # for cached properties and values in this instance
        self.__cache__ = {}
        self._calibrations = {}

        for name, trait in self._traits.items():
            trait.__init_owner_instance__(self)

            if trait.default is not Undefined:
                self.__cache__[name] = trait.default

    @util.hide_in_traceback
    def __init_subclass__(cls):
        cls._traits = dict(getattr(cls, "_traits", {}))
        cls._property_attrs = []
        cls._value_attrs = []
        cls._datareturn_attrs = []
        parent_traits = getattr(cls.__mro__[1], "_traits", {})

        # annotations = getattr(cls, '__annotations__', {})

        for name, trait in dict(cls._traits).items():
            # Apply the trait decorator to the object if it is "part 2" of a decorator
            obj = getattr(cls, name)

            if not isinstance(obj, Trait):
                if trait.role in (
                    Trait.ROLE_PROPERTY,
                    Trait.ROLE_DATARETURN,
                ) and callable(obj):
                    # if it's a method, decorate it
                    cls._traits[name] = trait(obj)
                else:
                    # if not decorating, clear from the traits dict, and emit a warning at runtime
                    thisclsname = cls.__qualname__
                    parentclsname = cls.__mro__[1].__qualname__
                    warn(
                        f"'{name}' in {thisclsname} is not a trait, but replaces one in parent class {parentclsname}"
                    )
                    del cls._traits[name]

                    continue

            setattr(cls, name, cls._traits[name])

        if cls._traits in HasTraitsMeta.__cls_namespace__:
            HasTraitsMeta.__cls_namespace__.remove(cls._traits)

        # finalize trait setup
        for name, trait in dict(cls._traits).items():
            if not hasattr(trait, "__objclass__"):
                trait.__set_name__(cls, name)
            trait.__init_owner_subclass__(cls)

            if trait.role == Trait.ROLE_VALUE:
                cls._value_attrs.append(name)
            elif trait.role == Trait.ROLE_DATARETURN:
                cls._datareturn_attrs.append(name)
            elif trait.role == Trait.ROLE_PROPERTY:
                cls._property_attrs.append(name)

    @util.hide_in_traceback
    def __notify__(self, name, value, type, cache):
        old = self.__cache__.setdefault(name, Undefined)

        msg = dict(new=value, old=old, owner=self, name=name, type=type, cache=cache)

        for handler in self.__notify_list__.values():
            handler(dict(msg))

        self.__cache__[name] = value

    def set_key(self, key, value, name=None):
        """implement this in subclasses to use `key` to set a parameter value from the
        Device with self.backend.

        property traits defined with "key=" call this to set values
        in the backend.
        """

        clsname = self.__class__.__qualname__
        raise NotImplementedError(
            f"implement {clsname}.get_key for access to key/value parameters on the device"
        )

    def get_key(self, key, name=None):
        """implement this in subclasses to use `key` to retreive a parameter value from the
        Device with self.backend.

        property traits defined with "key=" call this to retrieve values
        from the backend.
        """
        clsname = self.__class__.__qualname__
        raise NotImplementedError(
            f"implement {clsname}.get_key for access key/value parameters on the device"
        )

    @util.hide_in_traceback
    def __get_value__(self, name):
        """Get value of a trait for this value traits instance

        Arguments:
            name: Name of the trait
        Returns:
            cached value, or the trait default if it has not yet been set
        """
        return self.__cache__[name]

    @util.hide_in_traceback
    def __set_value__(self, name, value):
        """Set value of a trait for this value traits instance

        Arguments:
            name: Name of the trait
            value: value to assign
        Returns:
            None
        """
        # assignment to to self.__cache__ here would corrupt 'old' message key in __notify__
        pass


class Any(Trait, type=None):
    """allows any value"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        return value

    @util.hide_in_traceback
    def to_pythonic(self, value):
        return value


Trait.__annotations__["key"] = Any


def observe(obj, handler, name=Any, type_=("get", "set")):
    """Register a handler function to be called whenever a trait changes.

    The handler function takes a single message argument. This
    dictionary message has the keys

    * `new`: the updated value
    * `old`: the previous value
    * `owner`: the object that owns the trait
    * `name`: the name of the trait
    * 'event': 'set' or 'get'

    Arguments:
        handler: the handler function to call when the value changes
        names: notify only changes to these trait names (None to disable filtering)
    """

    def validate_name(n):
        attr = getattr(type(obj), n, Undefined)
        if attr is Undefined:
            raise TypeError(f'there is no attribute "{n}" to observe in "{obj}"')
        elif not isinstance(attr, Trait):
            raise TypeError(f"cannot observe {obj}.{n} because it is not a trait")

    if not callable(handler):
        raise ValueError(
            f"argument 'handler' is {repr(handler)}, which is not a callable"
        )

    if isinstance(name, str):
        validate_name(name)
        name = (name,)
    elif isinstance(name, (tuple, list)):
        for n in name:
            validate_name(n)
    elif name is not Any:
        raise ValueError(
            f"name argument {name} has invalid type - must be one of (str, tuple, list), or the value Any"
        )

    if isinstance(type, str):
        type_ = (type_,)

    def wrapped(msg):
        # filter according to name and type
        if name is not Any and msg["name"] not in name:
            return
        elif msg["type"] not in type_:
            return
        elif isinstance(msg["new"], Trait):
            raise TypeError(f"Trait instance returned as a callback value")
        handler(msg)

    if isinstance(obj, HasTraits):
        obj.__notify_list__[handler] = wrapped
    else:
        raise TypeError("object to observe must be an instance of Device")


def unobserve(obj, handler):
    """Unregister a handler function from notifications in obj."""
    if isinstance(obj, HasTraits):
        try:
            del obj.__notify_list__[handler]
        except KeyError as e:
            ex = e
        else:
            ex = None
        if ex:
            raise ValueError(f"{handler} was not registered to observe {obj}")
    else:
        raise TypeError("object to unobserve must be an instance of Device")


def find_trait_in_mro(cls):
    if issubclass(cls, DependentTrait):
        return find_trait_in_mro(type(cls._trait_dependencies["base"]))
    else:
        return cls


class DependentTrait(Trait):
    _trait_dependencies = set()

    def __set_name__(self, owner_cls, name):
        super().__set_name__(owner_cls, name)

        # propagate ownership of dependent traits, if available
        if isinstance(owner_cls, HasTraits):
            objclass = owner_cls
        elif hasattr(self, "__objclass__"):
            objclass = self.__objclass__
        else:
            return

        for trait in self._trait_dependencies.values():
            trait.__objclass__ = objclass

    def _validate_trait_dependencies(self, owner, allow_none: bool, operation="access"):
        if allow_none:
            return

        none_names = [
            f"{owner}.{trait.name}"
            for trait in self._trait_dependencies.values()
            if getattr(owner, trait.name) is None
        ]

        if len(none_names) == 1:
            raise ValueError(
                f"cannot {operation} {owner}.{self.name} while {none_names[0]} is None"
            )
        elif len(none_names) > 1:
            raise ValueError(
                f"cannot {operation} {owner}.{self.name} while {tuple(none_names)} are None"
            )

    @classmethod
    def derive(mixin_cls, template_trait, dependent_traits={}, *init_args, **init_kws):
        name = template_trait.__class__.__name__
        name = ("" if name.startswith("dependent_") else "dependent_") + name

        dependent_traits["base"] = template_trait

        traits_dict = {}

        for c in mixin_cls.__mro__:
            if issubclass(c, DependentTrait):
                traits_dict.update(c._trait_dependencies)

        traits_dict.update(dependent_traits)

        ns = dict(_trait_dependencies=traits_dict, **dependent_traits)

        ttype = type(name, (mixin_cls, find_trait_in_mro(type(template_trait))), ns)

        obj = ttype(*init_args, **init_kws)
        return obj


class RemappingCorrectionMixIn(DependentTrait):
    """act as another BoundedNumber trait calibrated with a mapping"""

    mapping: Any = None  # really a pandas Series

    EMPTY_STORE = dict(by_cal=None, by_uncal=None)

    def _min(self, owner):
        by_uncal = owner._calibrations.get(self.name, {}).get("by_uncal", None)
        if by_uncal is None:
            return None
        else:
            return by_uncal.min()

    def _max(self, owner):
        by_uncal = owner._calibrations.get(self.name, {}).get("by_uncal", None)
        if by_uncal is None:
            return None
        else:
            return by_uncal.max()

    def __init_owner_instance__(self, owner):
        self.set_mapping(self.mapping, owner=owner)
        observe(
            owner,
            self._on_base_trait_change,
            name=self._trait_dependencies["base"].name,
        )

    def _on_base_trait_change(self, msg):
        owner = msg["owner"]
        owner.__notify__(
            self.name,
            self.lookup_cal(msg["new"], owner),
            msg["type"],
            cache=msg["cache"],
        )

    def lookup_cal(self, uncal, owner):
        """look up and return the calibrated value, given the uncalibrated value"""
        owner_cal = owner._calibrations.get(self.name, self.EMPTY_STORE)
        if owner_cal.get("by_uncal", None) is None:
            return None

        try:
            return owner_cal["by_uncal"].loc[uncal]
        except KeyError:
            # spare us pandas details in the traceback
            util.logger.warning(
                f"{self.__repr__(owner_inst=owner)} has no entry at {repr(uncal)} {self.label}"
            )

        return None

    def find_uncal(self, cal, owner):
        """look up the calibrated value for the given uncalibrated value. In the event of a lookup
        error, then if `self.allow_none` evaluates as True, triggers return of None, or if
         `self.allow_none` evaluates False, ValueError is raised.
        """
        owner_cal = owner._calibrations.get(self.name, self.EMPTY_STORE)

        if owner_cal["by_uncal"] is None:
            return None

        i = owner_cal["by_cal"].index.get_loc(cal, method="nearest")
        return owner_cal["by_cal"].iloc[i]

    def set_mapping(self, series_or_uncal, cal=None, owner=None):
        """set the lookup mapping as `set_mapping(series)`, where `series` is a pandas Series (uncalibrated
        values in the index), or `set_mapping(cal_vector, uncal_vector)`, where both vectors have 1
        dimension of the same length.
        """

        if owner is None:
            raise ValueError(f"must pass owner to set_mapping")

        import pandas as pd

        if isinstance(series_or_uncal, pd.Series):
            by_uncal = pd.Series(series_or_uncal).copy()
        elif cal is not None:
            by_uncal = pd.Series(cal, index=series_or_uncal)
        elif series_or_uncal is None:
            return
        else:
            raise ValueError(
                f"must call set_mapping with None, a Series, or a pair of vector "
                f"arguments, not {series_or_uncal}"
            )
        by_uncal = by_uncal[~by_uncal.index.duplicated(keep="first")].sort_index()
        by_uncal.index.name = "uncal"
        by_uncal.name = "cal"

        by_cal = pd.Series(by_uncal.index, index=by_uncal.values, name="uncal")
        by_cal = by_cal[~by_cal.index.duplicated(keep="first")].sort_index()
        by_cal.index.name = "cal"

        owner._calibrations.setdefault(self.name, {}).update(
            by_cal=by_cal, by_uncal=by_uncal
        )

    @util.hide_in_traceback
    def __get__(self, owner, owner_cls=None):
        if owner is None or owner_cls is not self.__objclass__:
            return self

        # by_cal, by_uncal = owner._calibrations.get(self.name, (None, None))
        self._validate_trait_dependencies(owner, self.allow_none, "get")

        uncal = self._trait_dependencies["base"].__get__(owner, owner_cls)

        cal = self.lookup_cal(uncal, owner)

        if cal is None:
            ret = uncal
        else:
            ret = cal

        if hasattr(self, "name"):
            owner.__notify__(
                self.name,
                ret,
                "get",
                cache=self.cache or (self.role == self.ROLE_VALUE),
            )

        return ret

    @util.hide_in_traceback
    def __set__(self, owner, cal):
        owner_cal = owner._calibrations.get(self.name, self.EMPTY_STORE)
        self._validate_trait_dependencies(owner, False, "set")

        # start with type conversion and validation on the requested calibrated value
        cal = self._trait_dependencies["base"].to_pythonic(cal)

        # lookup the uncalibrated value that results in the nearest calibrated result
        uncal = self.find_uncal(cal, owner)

        if uncal is None:
            self._trait_dependencies["base"].__set__(owner, cal)
        elif uncal != type(self._trait_dependencies["base"]).validate(
            self, uncal, owner
        ):
            # raise an exception if the calibration table contains invalid
            # values, instead
            raise ValueError(
                f"calibration lookup in {self.__repr__(owner_inst=owner)} produced invalid value {repr(uncal)}"
            )
        else:
            # execute the set
            self._trait_dependencies["base"].__set__(owner, uncal)

        if hasattr(self, "name"):
            owner.__notify__(
                self.name,
                cal,
                "set",
                cache=self.cache or (self.role == self.ROLE_VALUE),
            )


class TableCorrectionMixIn(RemappingCorrectionMixIn):
    _CAL_TABLE_KEY = "table"

    path_trait = None  # a dependent Unicode trait

    index_lookup_trait = None  # a dependent trait

    table_index_column: str = None

    def __init_owner_instance__(self, owner):
        super().__init_owner_instance__(owner)

        observe(
            owner,
            self._on_cal_update_event,
            name=[self.path_trait.name, self.index_lookup_trait.name],
            type_="set",
        )

    def _on_cal_update_event(self, msg):
        owner = msg["owner"]

        if msg["name"] == self.path_trait.name:
            # if msg['new'] == msg['old']:
            #     return

            path = msg["new"]
            index = getattr(owner, self.index_lookup_trait.name)

            ret = self._load_calibration_table(owner, path)
            self._update_index_value(owner, index)

            return ret

        elif msg["name"] == self.index_lookup_trait.name:
            # if msg['new'] == msg['old']:
            #     return
            path = getattr(owner, self.path_trait.name)
            index = msg["new"]

            if self._CAL_TABLE_KEY not in owner._calibrations.get(self.name, {}):
                self._load_calibration_table(owner, path)

            ret = self._update_index_value(owner, index)

            return ret

        else:
            raise KeyError(f"unsupported trait name {msg['name']}")

        # return self._update_index_value(msg["owner"], msg["new"])

    def _load_calibration_table(self, owner, path):
        """ stash the calibration table from disk
        """
        import pandas as pd
        from pathlib import Path

        def read(path):
            # quick read
            cal = pd.read_csv(str(path), index_col=self.table_index_column, dtype=float)

            cal.columns = cal.columns.astype(float)
            if self.index_lookup_trait.max in cal.index:
                cal.drop(self.index_lookup_trait.max, axis=0, inplace=True)
            #    self._cal_offset.values[:] = self._cal_offset.values-self._cal_offset.columns.values[np.newaxis,:]

            owner._calibrations.setdefault(self.name, {}).update(
                {self._CAL_TABLE_KEY: cal}
            )

            owner._logger.debug(f"calibration data read from {path}")

        if path is None:
            if not self.allow_none:
                raise ValueError(
                    f"{self} defined with allow_none=False; path_trait must not be None"
                )
            else:
                return None

        read(path)

    def _touch_table(self, owner):
        # make sure that calibrations have been initialized
        table = owner._calibrations.get(self.name, {}).get(self._CAL_TABLE_KEY, None)

        if table is None:
            path = getattr(owner, self.path_trait.name)
            index = getattr(owner, self.index_lookup_trait.name)

            if None not in (path, index):
                setattr(owner, self.path_trait.name, path)
                setattr(owner, self.index_lookup_trait.name, index)

    def _update_index_value(self, owner, index_value):
        """update the calibration on change of index_value"""
        cal = owner._calibrations.get(self.name, {}).get(self._CAL_TABLE_KEY, None)

        if cal is None:
            txt = f"index_value change has no effect because calibration_data has not been set"
        elif index_value is None:
            cal = None
            txt = f"set {owner}.{self.index_lookup_trait.name} to enable calibration"
        else:
            # pull in the calibration mapping specific to this index_value
            i_freq = cal.index.get_loc(index_value, "nearest")
            cal = cal.iloc[i_freq]
            txt = f"calibrated at {index_value/1e6:0.3f} MHz"

        self.set_mapping(cal, owner=owner)

    @util.hide_in_traceback
    def __get__(self, owner, owner_cls=None):
        if owner is None or owner_cls is not self.__objclass__:
            return self

        self._touch_table(owner)

        return super().__get__(owner, owner_cls)

    @util.hide_in_traceback
    def __set__(self, owner, cal):
        self._touch_table(owner)
        super().__set__(owner, cal)



class TransformMixIn(DependentTrait):
    """act as an arbitrarily-defined (but reversible) transformation of another BoundedNumber trait"""

    _forward: Any = lambda x, y: x
    _reverse: Any = lambda x, y: x

    def __init_owner_instance__(self, owner):
        super().__init_owner_instance__(owner)
        observe(owner, self.__owner_event__)

    def __owner_event__(self, msg):
        # pass on a corresponding notification when self._trait_dependencies['base'] changes
        base_trait = self._trait_dependencies["base"]

        if msg["name"] != getattr(base_trait, "name", None) or not hasattr(
            base_trait, "__objclass__"
        ):
            return

        owner = msg["owner"]
        owner.__notify__(self.name, msg["new"], msg["type"], cache=msg["cache"])

    def _transformed_extrema(self, owner):
        base_trait = self._trait_dependencies["base"]
        base_bounds = [base_trait._min(owner), base_trait._max(owner)]

        other_trait = self._trait_dependencies.get("other", None)

        if other_trait is None:
            trial_bounds = [
                self._forward(base_bounds[0]),
                self._forward(base_bounds[1]),
            ]
        else:
            other_value = getattr(owner, other_trait.name)
            # other_bounds = [
            #     other_trait._min(owner),
            #     other_trait._max(owner),
            # ]

            # trial_bounds = [
            #     self._forward(base_bounds[0], other_bounds[0]),
            #     self._forward(base_bounds[0], other_bounds[1]),
            #     self._forward(base_bounds[1], other_bounds[0]),
            #     self._forward(base_bounds[1], other_bounds[1]),
            # ]
            trial_bounds = [
                self._forward(base_bounds[0], other_value),
                self._forward(base_bounds[1], other_value),
            ]

        if None in trial_bounds:
            return None, None

        return min(trial_bounds), max(trial_bounds)

    def _min(self, owner):
        # TODO: ensure this works properly for any reversible self._forward()?
        lo, hi = self._transformed_extrema(owner)

        if lo is None:
            return None
        else:
            return min(lo, hi)

    def _max(self, owner):
        # TODO: ensure this works properly for any reversible self._forward()?
        lo, hi = self._transformed_extrema(owner)

        if hi is None:
            return None
        else:
            return max(lo, hi)

    def __get__(self, owner, owner_cls=None):
        if owner is None or owner_cls is not self.__objclass__:
            return self

        base_value = self._trait_dependencies["base"].__get__(owner, owner_cls)

        if "other" in self._trait_dependencies:
            other_value = self._trait_dependencies["other"].__get__(owner, owner_cls)
            ret = self._forward(base_value, other_value)
        else:
            ret = self._forward(base_value)

        if hasattr(self, "name"):
            owner.__notify__(
                self.name,
                ret,
                "get",
                cache=self.cache or (self.role == self.ROLE_VALUE),
            )

        return ret

    def __set__(self, owner, value_request):
        # use the other to the value into the proper format and validate it
        base_trait = self._trait_dependencies["base"]
        value = base_trait.to_pythonic(value_request)

        # now reverse the transformation
        if "other" in self._trait_dependencies:
            other_trait = self._trait_dependencies["other"]
            other_value = other_trait.__get__(owner, other_trait.__objclass__)

            base_value = self._reverse(value, other_value)
        else:
            base_value = self._reverse(value)

        # set the value of the base trait with the reverse-transformed value
        base_trait.__set__(owner, base_value)

        if hasattr(self, "name"):
            owner.__notify__(
                self.name,
                value,
                "set",
                cache=self.cache or (self.role == self.ROLE_VALUE),
            )


class BoundedNumber(Trait):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""

    default: ThisType = None
    allow_none: bool = True
    min: ThisType = None
    max: ThisType = None

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not isinstance(value, (bytes, str, bool, numbers.Number)):
            raise ValueError(
                f"a '{type(self).__qualname__}' trait value must be a numerical, str, or bytes instance"
            )

        # Check bounds once it's a numerical type
        min = self._min(owner)
        max = self._max(owner)

        if max is not None and value > max:
            raise ValueError(
                f"{value} is greater than the max limit {max} of {self._owned_name(owner)}"
            )
        if min is not None and value < min:
            raise ValueError(
                f"{value} is less than the min limit {min} of {self._owned_name(owner)}"
            )

        return value

    def _max(self, owner):
        """overload this to dynamically compute max"""
        return self.max

    def _min(self, owner):
        """overload this to dynamically compute max"""
        return self.min

    path_trait: Any = None  # TODO: should be a Unicode string trait

    index_lookup_trait: Any = None  # TODO: this is a trait that should almost certainly be a BoundedNumber

    table_index_column: str = None

    def calibrate_from_table(
        self,
        path_trait,
        index_lookup_trait,
        *,
        table_index_column: str = None,
        help="",
        label=Undefined,
        allow_none=False,
    ):
        """generate a new Trait with value dependent on another trait. their configuration
        comes from a trait in the owner.

        Arguments:
            offset_name: the name of a value trait in the owner containing a numerical offset
            lookup1d: a table containing calibration data, or None to configure later
        """

        if label is Undefined:
            label = self.label

        ret = TableCorrectionMixIn.derive(
            self,
            dict(path_trait=path_trait, index_lookup_trait=index_lookup_trait,),
            help=help,
            label=self.label if label is Undefined else label,
            sets=self.sets,
            gets=self.gets,
            allow_none=allow_none,
            table_index_column=table_index_column,
        )

        return ret

    def calibrate_from_expression(
        self,
        trait_expression,
        help: str = "",
        label: str = Undefined,
        allow_none: bool = False,
    ):
        if isinstance(self, DependentTrait):
            # This a little unsatisfying, but the alternative would mean
            # solving the trait_expression for the trait `self`
            obj = trait_expression
            while isinstance(obj, DependentTrait):
                obj = obj._trait_dependencies["base"]
                if obj == self:
                    break
            else:
                raise TypeError(
                    f"the trait being calibrated must also be the first trait in the calibration expression"
                )

        return self.update(
            trait_expression, help=help, label=label, allow_none=allow_none
        )

    # def calibrate(
    #     self,
    #     offset=Undefined,
    #     mapping=Undefined,
    #     table=Undefined,
    #     help="",
    #     label=Undefined,
    #     allow_none=False,
    # ):
    #     """generate a new Trait with value dependent on another trait. their configuration
    #     comes from a trait in the owner.

    #     Arguments:
    #         offset_name: the name of a value trait in the owner containing a numerical offset
    #         lookup1d: a table containing calibration data, or None to configure later
    #     """

    #     params = {}
    #     if mapping is not Undefined:
    #         mixin = RemappingCorrectionMixIn
    #         params["mapping"] = mapping

    #     elif offset is not Undefined:
    #         mixin = OffsetCorrectionMixIn
    #         params["offset"] = offset

    #     if label is Undefined:
    #         label = self.label

    #     if len(params) != 1:
    #         raise ValueError(f"must set exactly one of `offset`, `lookup1d`, and `lookup2d`")

    #     return mixin.derive(
    #         self,
    #         help=help,
    #         label=self.label,
    #         sets=self.sets,
    #         gets=self.gets,
    #         allow_none=allow_none,
    #         **params,
    #     )

    def transform(
        self,
        other_trait: Trait,
        forward: callable,
        reverse: callable,
        help="",
        allow_none=False,
    ):
        """generate a new Trait subclass that adjusts values in other traits.

        Arguments:
            forward: implementation of the forward transformation
            reverse: implementation of the reverse transformation
        """

        obj = TransformMixIn.derive(
            self,
            dependent_traits={} if other_trait is None else dict(other=other_trait),
            help=help,
            label=self.label,
            sets=self.sets,
            gets=self.gets,
            allow_none=allow_none,
            _forward=forward,
            _reverse=reverse,
        )

        return obj

    def __neg__(self):
        def neg(x, y=None):
            return None if x is None else -x

        return self.transform(
            None, neg, neg, allow_none=self.allow_none, help=f"-1*({self.help})"
        )

    def __add__(self, other):
        def add(x, y):
            return None if None in (x, y) else x + y

        def sub(x, y):
            return None if None in (x, y) else x - y

        return self.transform(
            other, add, sub, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )

    __radd__ = __add__

    def __sub__(self, other):
        def add(x, y):
            return None if None in (x, y) else x + y

        def sub(x, y):
            return None if None in (x, y) else x - y

        return self.transform(
            other, sub, add, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )

    def __rsub__(self, other):
        def add(x, y):
            return None if None in (x, y) else y + x

        def sub(x, y):
            return None if None in (x, y) else y - x

        return self.transform(
            other, sub, add, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )

    def __mul__(self, other):
        def mul(x, y):
            return None if None in (x, y) else x * y

        def div(x, y):
            return None if None in (x, y) else x / y

        return self.transform(
            other, mul, div, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )

    __rmul__ = __mul__

    def __truediv__(self, other):
        def mul(x, y):
            return None if None in (x, y) else x * y

        def div(x, y):
            return None if None in (x, y) else x / y

        return self.transform(
            other, div, mul, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )

    def __rdiv__(self, other):
        def mul(x, y):
            return None if None in (x, y) else y * x

        def div(x, y):
            return None if None in (x, y) else y / x

        return self.transform(
            other, div, mul, allow_none=self.allow_none, help=f"({self.help}) + {other}"
        )


class NonScalar(Any):
    """generically non-scalar data, such as a list, array, but not including a string or bytes"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if isinstance(value, (bytes, str)):
            raise ValueError(f"given text data but expected a non-scalar data")
        if not hasattr(value, "__iter__") and not hasattr(value, "__len__"):
            raise ValueError(f"expected non-scalar data but given a non-iterable")
        return value


class Int(BoundedNumber, type=int):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""


class Float(BoundedNumber, type=float):
    """accepts numerical, str, or bytes values, following normal python casting procedures (with bounds checking)"""

    step: ThisType = None

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        value = super().validate(value, owner)
        if self.step is not None:
            mod = value % self.step
            if mod < self.step / 2:
                return value - mod
            else:
                return value - (mod - self.step)
        return value


class Complex(Trait, type=complex):
    """accepts numerical or str values, following normal python casting procedures (with bounds checking)"""

    allow_none: bool = False


class Bool(Trait, type=bool):
    """accepts boolean or numeric values, or a case-insensitive match to one of ('true',b'true','false',b'false')"""

    allow_none: bool = False

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if isinstance(value, (bool, numbers.Number)):
            return value
        elif isinstance(value, (str, bytes)):
            lvalue = value.lower()
            if lvalue in ("true", b"true"):
                return True
            elif lvalue in ("false", b"false"):
                return False
        raise ValueError(
            f"'{self.__repr__(owner_inst=owner)}' accepts only boolean, numerical values,"
            "or one of ('true',b'true','false',b'false'), case-insensitive"
        )


class String(Trait):
    """base class for string types, which adds support for case sensitivity arguments"""

    case: bool = True
    # allow_none: bool = True # let's not override this default

    @util.hide_in_traceback
    def contains(self, iterable, value):
        if not self.case:
            iterable = [v.lower() for v in iterable]
            value = value.lower()
        return value in iterable


class Unicode(String, type=str):
    """accepts strings or numeric values only; convert others explicitly before assignment"""

    default: ThisType = ""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not isinstance(value, (str, numbers.Number)):
            raise ValueError(
                f"'{type(self).__qualname__}' traits accept values of str or numerical type, not {type(value).__name__}"
            )
        return value


class Bytes(String, type=bytes):
    """accepts bytes objects only - encode str (unicode) explicitly before assignment"""

    default: ThisType = b""


class Iterable(Trait):
    """accepts any iterable"""

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        if not hasattr(value, "__iter__"):
            raise ValueError(
                f"'{type(self).__qualname__}' traits accept only iterable values"
            )
        return value


class Dict(Iterable, type=dict):
    """accepts any type of iterable value accepted by python `dict()`"""


class List(Iterable, type=list):
    """accepts any type of iterable value accepted by python `list()`"""


class Tuple(Iterable, type=tuple):
    """accepts any type of iterable value accepted by python `tuple()`"""

    sets: bool = False


class Path(Trait, type=Path):
    must_exist: bool = False
    """ does the path need to exist when set? """

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        path = self.type(value)

        if self.must_exist and not path.exists():
            raise IOError()

        return path


# class PandasDataFrame(NonScalar, type=pd.DataFrame):
#     pass


# class PandasSeries(NonScalar, type=pd.Series):
#     pass


# class NumpyArray(NonScalar, type=np.ndarray):
#     pass


class NetworkAddress(Unicode):
    """a IDN-compatible network address string, such as an IP address or DNS hostname"""

    accept_port: bool = True

    @util.hide_in_traceback
    def validate(self, value, owner=None):
        """Rough IDN compatible domain validator"""

        host, *extra = value.split(":", 1)

        if len(extra) > 0:
            port = extra[0]
            try:
                int(port)
            except ValueError:
                raise ValueError(f'port {port} in "{value}" is invalid')

            if not self.accept_port:
                raise ValueError(
                    f"{self} does not accept a port number (accept_port=False)"
                )

        for validate in _val.ipv4, _val.ipv6, _val.domain, _val.slug:
            if validate(host):
                break
        else:
            raise ValueError("invalid host address")

        return value


VALID_TRAIT_ROLES = Trait.ROLE_VALUE, Trait.ROLE_PROPERTY, Trait.ROLE_DATARETURN


def subclass_namespace_traits(namespace_dict, role, omit_trait_attrs):
    for name, attr in dict(namespace_dict).items():
        if isclass(attr) and issubclass(attr, Trait):
            # subclass our traits with the given role
            new_trait = type(name, (attr,), dict(role=role))
            new_trait.role = role

            # clean out annotations for stub generation
            new_trait.__annotations__ = dict(new_trait.__annotations__)
            for drop_attr in omit_trait_attrs:
                new_trait.__annotations__.pop(drop_attr)
            new_trait.__module__ = namespace_dict["__name__"]

            namespace_dict[name] = new_trait



import typing
from . import util as util
from _typeshed import Incomplete
from collections.abc import Generator
Undefined: Incomplete
T: Incomplete


class ThisType():
    ...


class HasTraitsMeta(type):
    __cls_namespace__: Incomplete

    @classmethod
    def __prepare__(cls, names, bases, **kws):
        ...


class Trait():

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    ROLE_VALUE: str
    ROLE_PROPERTY: str
    ROLE_DATARETURN: str
    ROLE_UNSET: str
    type: Incomplete
    role: Incomplete
    default: ThisType
    key: Undefined
    func: typing.Callable
    help: str
    label: str
    sets: bool
    gets: bool
    cache: bool
    only: tuple
    allow_none: bool
    remap: dict
    kws: Incomplete
    metadata: Incomplete
    remap_inbound: Incomplete

    @classmethod
    def __init_subclass__(cls, type=...) -> None:
        ...

    def copy(self, new_type: Incomplete | None=..., **update_kws):
        ...
    __objclass__: Incomplete
    name: Incomplete

    def __set_name__(self, owner_cls, name) -> None:
        ...

    def __init_owner_subclass__(self, owner_cls) -> None:
        ...

    def __init_owner_instance__(self, owner) -> None:
        ...

    def __set__(self, owner, value) -> None:
        ...

    def __get__(self, owner, owner_cls: Incomplete | None=...):
        ...

    def __cast_get__(self, owner, value, strict: bool=...):
        ...

    def to_pythonic(self, value):
        ...

    def from_pythonic(self, value):
        ...

    def validate(self, value, owner: Incomplete | None=...):
        ...

    def contains(self, iterable, value):
        ...

    def __call__(self, func):
        ...

    def doc(self):
        ...

    def doc_params(self, omit=...):
        ...

    def update(self, obj: Incomplete | None=..., **attrs):
        ...

def hold_trait_notifications(owner) -> Generator[None, None, None]:
    ...


class HasTraits(metaclass=HasTraitsMeta):
    __notify_list__: Incomplete
    __cls_namespace__: Incomplete
    __cache__: Incomplete

    def __init__(self, **values) -> None:
        ...

    def __init_subclass__(cls) -> None:
        ...

    def __notify__(self, name, value, type, cache) -> None:
        ...

    def set_key(self, key, value, name: Incomplete | None=...) -> None:
        ...

    def get_key(self, key, name: Incomplete | None=...) -> None:
        ...

    def __get_value__(self, name):
        ...

    def __set_value__(self, name, value) -> None:
        ...


class Any(Trait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...

    def validate(self, value, owner: Incomplete | None=...):
        ...

    def to_pythonic(self, value):
        ...

def observe(obj, handler, name=..., type_=...) -> None:
    ...

def unobserve(obj, handler) -> None:
    ...

def find_trait_in_mro(cls):
    ...


class DependentTrait(Trait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...

    def __set_name__(self, owner_cls, name) -> None:
        ...

    @classmethod
    def derive(mixin_cls, template_trait, dependent_traits=..., *init_args, **init_kws):
        ...


class RemappingCorrectionMixIn(DependentTrait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        mapping=None
    ):
        ...
    mapping: Any
    EMPTY_STORE: Incomplete

    def __init_owner_instance__(self, owner) -> None:
        ...

    def lookup_cal(self, uncal, owner):
        ...

    def find_uncal(self, cal, owner):
        ...

    def set_mapping(self, series_or_uncal, cal: Incomplete | None=..., owner: Incomplete | None=...) -> None:
        ...

    def __get__(self, owner, owner_cls: Incomplete | None=...):
        ...

    def __set__(self, owner, cal) -> None:
        ...


class TableCorrectionMixIn(RemappingCorrectionMixIn):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        mapping=None,
        table_index_column: str=None
    ):
        ...
    path_trait: Incomplete
    index_lookup_trait: Incomplete
    table_index_column: str

    def __init_owner_instance__(self, owner) -> None:
        ...

    def __get__(self, owner, owner_cls: Incomplete | None=...):
        ...

    def __set__(self, owner, cal) -> None:
        ...


class TransformMixIn(DependentTrait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...

    def __init_owner_instance__(self, owner) -> None:
        ...

    def __owner_event__(self, msg) -> None:
        ...

    def __get__(self, owner, owner_cls: Incomplete | None=...):
        ...

    def __set__(self, owner, value_request) -> None:
        ...


class BoundedNumber(Trait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=True,
        remap: dict={},
        min=None,
        max=None,
        path_trait=None,
        index_lookup_trait=None,
        table_index_column: str=None
    ):
        ...
    default: ThisType
    allow_none: bool
    min: ThisType
    max: ThisType

    def validate(self, value, owner: Incomplete | None=...):
        ...
    path_trait: Any
    index_lookup_trait: Any
    table_index_column: str

    def calibrate_from_table(
        self,
        path_trait,
        index_lookup_trait,
        *,
        table_index_column: str=...,
        help: str=...,
        label=...,
        allow_none: bool=...
    ):
        ...

    def calibrate_from_expression(
        self,
        trait_expression,
        help: str=...,
        label: str=...,
        allow_none: bool=...
    ):
        ...

    def transform(
        self,
        other_trait: Trait,
        forward: callable,
        reverse: callable,
        help: str=...,
        allow_none: bool=...
    ):
        ...

    def __neg__(self):
        ...

    def __add__(self, other):
        ...
    __radd__: Incomplete

    def __sub__(self, other):
        ...

    def __rsub__(self, other):
        ...

    def __mul__(self, other):
        ...
    __rmul__: Incomplete

    def __truediv__(self, other):
        ...

    def __rdiv__(self, other):
        ...


class NonScalar(Any):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...

    def validate(self, value, owner: Incomplete | None=...):
        ...


class Int(BoundedNumber):

    def __init__(
        default: int=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=True,
        remap: dict={},
        min: int=None,
        max: int=None,
        path_trait=None,
        index_lookup_trait=None,
        table_index_column: str=None
    ):
        ...
    ...


class Float(BoundedNumber):

    def __init__(
        default: float=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=True,
        remap: dict={},
        min: float=None,
        max: float=None,
        path_trait=None,
        index_lookup_trait=None,
        table_index_column: str=None,
        step: float=None
    ):
        ...
    step: ThisType

    def validate(self, value, owner: Incomplete | None=...):
        ...


class Complex(Trait):

    def __init__(
        default: complex=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    allow_none: bool


class Bool(Trait):

    def __init__(
        default: bool=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    allow_none: bool

    def validate(self, value, owner: Incomplete | None=...):
        ...


class String(Trait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        case: bool=True
    ):
        ...
    case: bool

    def contains(self, iterable, value):
        ...


class Unicode(String):

    def __init__(
        default: str='',
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        case: bool=True
    ):
        ...
    default: ThisType

    def validate(self, value, owner: Incomplete | None=...):
        ...


class Bytes(String):

    def __init__(
        default: bytes=b'',
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        case: bool=True
    ):
        ...
    default: ThisType


class Iterable(Trait):

    def __init__(
        default=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...

    def validate(self, value, owner: Incomplete | None=...):
        ...


class Dict(Iterable):

    def __init__(
        default: dict=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    ...


class List(Iterable):

    def __init__(
        default: list=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    ...


class Tuple(Iterable):

    def __init__(
        default: tuple=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=False,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={}
    ):
        ...
    sets: bool


class Path(Trait):

    def __init__(
        default: Path=None,
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        must_exist: bool=False
    ):
        ...
    must_exist: bool

    def validate(self, value, owner: Incomplete | None=...):
        ...


class NetworkAddress(Unicode):

    def __init__(
        default: str='',
        key=None,
        func: _CallableType=None,
        help: str='',
        label: str='',
        sets: bool=True,
        gets: bool=True,
        cache: bool=False,
        only: tuple=(),
        allow_none: bool=False,
        remap: dict={},
        case: bool=True,
        accept_port: bool=True
    ):
        ...
    accept_port: bool

    def validate(self, value, owner: Incomplete | None=...):
        ...
VALID_TRAIT_ROLES: Incomplete

def subclass_namespace_traits(namespace_dict, role, omit_trait_attrs) -> None:
    ...



import logging
from . import _device as core
from _typeshed import Incomplete


class LogStreamBuffer():

    def __init__(self) -> None:
        ...

    def write(self, msg):
        ...

    def read(self):
        ...

    def flush(self) -> None:
        ...


class LogStderr(core.Device):

    def __init__(self, resource: str='str'):
        ...
    log: str

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    def write(self, what) -> None:
        ...

    def flush(self) -> None:
        ...


class Email(core.Device):

    def __init__(
        self,
        resource: str='str',
        port: str='int',
        sender: str='str',
        recipients: str='list',
        success_message: str='str',
        failure_message: str='str'
    ):
        ...
    resource: Incomplete
    port: Incomplete
    sender: Incomplete
    recipients: Incomplete
    success_message: Incomplete
    failure_message: Incomplete
    backend: Incomplete

    def open(self) -> None:
        ...

    def close(self) -> None:
        ...

    def send_summary(self):
        ...


class JSONFormatter(logging.Formatter):
    t0: Incomplete

    def __init__(self, *args, **kws) -> None:
        ...

    @staticmethod
    def json_serialize_dates(obj):
        ...

    def format(self, rec):
        ...


class Host(core.Device):

    def __init__(self, resource: str='str', git_commit_in: str='NoneType'):
        ...
    git_commit_in: Incomplete
    time_format: str
    backend: Incomplete

    def open(self) -> None:
        ...

    @classmethod
    def __imports__(self) -> None:
        ...

    def close(self) -> None:
        ...

    def metadata(self):
        ...

    def time(self):
        ...

    def log(self):
        ...

    def git_commit_id(self):
        ...

    def git_remote_url(self):
        ...

    def hostname(self):
        ...

    def git_browse_url(self):
        ...

    def git_pending_changes(self):
        ...
